//--------------------------------------------------------------------------//
// xmtnimage3.cc                                                            //
// Image file reading, writing                                              //
// Latest revision: 10-04-2023                                              //
//--------------------------------------------------------------------------//
 
#include "xmtnimage.h"
extern Globals     g;
extern Image      *z;
extern int        ci;
extern int    blocked_while_saving;

int override=0;
int decrypt(char *filename);
int decompress(char *filename);
void delete_temporary_file(char *temp_filename);
void create_temporary_file(char *temp_filename, char *filename);

////  If a new format is added, increase this value
////  also make sure the identifier in xmtnimage.h is
////  within this range
////  The identifiers to show up in the list must be
////  in sequential order

const int STDFORMATS=14;
int noofformats;  
char identifier[64] = "NONE";
int hitnonstandard=0;
int combinedino;
static char read_filename[FILENAMELENGTH]="Untitled";
int in_save=0;
Dialog *in_save_dialog = 0;
int in_read=0;

//// want_color_type      0  1  2  3  4  5  6  7  8  9
int w_bpp[10]       = { 1, 8, 8,15,16,24,32,48, 0, 8 };
int w_noofcolors[10]= { 1, 1, 1, 3, 3, 3, 4, 3, 0, 1 };
int w_redbpp[10]    = { 0, 6, 6, 5, 5, 8, 8,16, 0, 6 };
int w_greenbpp[10]  = { 0, 6, 6, 5, 6, 8, 8,16, 0, 6 };
int w_bluebpp[10]   = { 0, 6, 6, 5, 5, 8, 8,16, 0, 6 };
int w_blackbpp[10]  = { 0, 0, 0, 0, 0, 0, 8, 0, 0, 6 };
int w_savecmyk[10]  = { 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 };

 
//--------------------------------------------------------------------------//
// read_start                                                               //
// Entry point for reading images                                           //
//--------------------------------------------------------------------------//
void readstart(int noofargs, char **arg)
{  
  if(memorylessthan(16384)){ message(g.nomemory,ERROR);  return; }
  static char **filelist = NULL;
  int j,k;
  char temp2[202];
  char *ptr; 
  static Dialog *dialog;
  if(in_read) return;
  in_read = 1;
  g.getout=0;
  if(noofargs==0)
  {  dialog = new Dialog;
     if(dialog==NULL){ message(g.nomemory); return; }
     strcpy(dialog->title,"Read Image");
     strcpy(dialog->radio[0][0],"File type");
     strcpy(dialog->radio[0][1],"Auto file type");
     strcpy(dialog->radio[0][2],"Raw bytes");
     strcpy(dialog->radio[0][3],"ASCII");
     strcpy(dialog->radio[0][4],"Raw 3-D");
     strcpy(dialog->radio[1][0],"Color reduction");
     strcpy(dialog->radio[1][1],"Quantization");
     strcpy(dialog->radio[1][2],"Fit current palette");
     strcpy(dialog->radio[1][3],"None");

  
     dialog->radioset[0] = g.read_raw + 1;
     dialog->radioset[1] = g.want_quantize;
     dialog->radiono[0] = 5;
     dialog->radiono[1] = 4;
     dialog->radiono[2] = 0;
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;

     strcpy(dialog->boxes[0],"File Parameters");
     strcpy(dialog->boxes[1],"Filename");
     strcpy(dialog->boxes[2],"X position");
     strcpy(dialog->boxes[3],"Y position");
     strcpy(dialog->boxes[4],"X size");
     strcpy(dialog->boxes[5],"Y size");
     strcpy(dialog->boxes[6],"CMYK->RGB");
     strcpy(dialog->boxes[7],"Convert->Gray scale");
     strcpy(dialog->boxes[8],"Convert->Scrn depth");
     strcpy(dialog->boxes[9],"Negative image");
     strcpy(dialog->boxes[10],"Separate window");
     strcpy(dialog->boxes[11],"Swap bytes");

     strcpy(dialog->boxes[12],"Pre-processing");

     strcpy(dialog->boxes[13],"Decompress file");
     strcpy(dialog->boxes[14],"Command");     // compression
     strcpy(dialog->boxes[15],"Extension");

     strcpy(dialog->boxes[16],"Decrypt file");// encryption
     strcpy(dialog->boxes[17],"Command");
     strcpy(dialog->boxes[18],"Extension");

     strcpy(dialog->boxes[19],"Split frames");
     strcpy(dialog->boxes[20],"Fix CMYK");

     dialog->boxtype[0]=LABEL;
     dialog->boxtype[1]=MULTIFILENAME; dialog->boxcount[1]=0;
     dialog->boxtype[2]=INTCLICKBOX; 
     dialog->boxtype[3]=INTCLICKBOX;
     dialog->boxtype[4]=INTCLICKBOX;
     dialog->boxtype[5]=INTCLICKBOX;
     dialog->boxtype[6]=TOGGLE;    
     dialog->boxtype[7]=TOGGLE;     
     dialog->boxtype[8]=TOGGLE;     
     dialog->boxtype[9]=TOGGLE;     
     dialog->boxtype[10]=TOGGLE;     
     dialog->boxtype[11]=TOGGLE;     
     dialog->boxtype[12]=LABEL;     
     dialog->boxtype[13]=TOGGLE;     
     dialog->boxtype[14]=STRING;     
     dialog->boxtype[15]=STRING;     
     dialog->boxtype[16]=TOGGLE;     
     dialog->boxtype[17]=STRING;     
     dialog->boxtype[18]=STRING;     
     dialog->boxtype[19]=TOGGLE;     
     dialog->boxtype[20]=TOGGLE;     

     dialog->boxmin[2]=0; dialog->boxmax[2]=g.xres;
     dialog->boxmin[3]=0; dialog->boxmax[3]=g.yres;
     dialog->boxmin[4]=0; dialog->boxmax[4]=100;
     dialog->boxmin[5]=0; dialog->boxmax[5]=100;

     strcpy(dialog->answer[1][0], read_filename);
     itoa(g.tif_xoffset,temp2,10);
     strcpy(dialog->answer[2][0], temp2);
     itoa(g.tif_yoffset,temp2,10);
     strcpy(dialog->answer[3][0], temp2);
     doubletostring(g.tif_xsize*100,5,temp2);       //  tif_xsize and tif_ysize are
                                          //  temporarily multiplied by 100
                                          //  for use in click box.
     strcpy(dialog->answer[4][0], temp2);
     doubletostring(g.tif_ysize*100,5,temp2);
     strcpy(dialog->answer[5][0], temp2);
     if(g.read_cmyk)     dialog->boxset[6]=1; else dialog->boxset[6]=0;
     if(g.read_grayscale)dialog->boxset[7]=1; else dialog->boxset[7]=0;
     if(g.read_convert)  dialog->boxset[8]=1; else dialog->boxset[8]=0;
     if(g.tif_positive)  dialog->boxset[9]=0; else dialog->boxset[9]=1;
     if(g.want_shell)    dialog->boxset[10]=1; else dialog->boxset[10]=0;
     if(g.read_swap_bytes) dialog->boxset[11]=1; else dialog->boxset[11]=0;

     dialog->boxset[13] = g.want_decompression;
     strcpy(dialog->answer[14][0], g.decompression);
     strcpy(dialog->answer[15][0], g.compression_ext);

     dialog->boxset[16] = g.want_decryption;
     strcpy(dialog->answer[17][0], g.decryption);
     strcpy(dialog->answer[18][0], g.encryption_ext);

     dialog->boxset[19] = g.read_split_frames;
     dialog->boxset[20] = g.fix_cmyk;

     dialog->boxlist[1] = filelist;
     dialog->noofradios=2;
     dialog->noofboxes=20;
     dialog->helptopic=11;  
     dialog->want_changecicb = 1;
     dialog->f1 = readcheck;
     dialog->f2 = null;
     dialog->f3 = null;
     dialog->f4 = readfinish;
     dialog->f5 = null;
     dialog->f6 = null;
     dialog->f7 = null;
     dialog->f8 = null;
     dialog->f9 = null;
     dialog->width = 0;  // calculate automatically
     dialog->height = 0; // calculate automatically
     dialog->transient = 1;
     dialog->wantraise = 0;
     dialog->radiousexy = 0;
     dialog->boxusexy = 0;
     dialog->radioextrawidth = 0;
     strcpy(dialog->path,".");
     dialog->message[0] = 0;      
     dialog->busy = 0;
     dialogbox(dialog);

  }else
  {  if(noofargs>1) g.tif_xoffset = atoi(arg[2]);
     if(noofargs>2) g.tif_yoffset = atoi(arg[3]);
     if(noofargs>3) g.tif_xsize   = strtod(arg[4],&ptr);
     if(noofargs>4) g.tif_ysize   = strtod(arg[5],&ptr);
     if(noofargs>0) readimage(arg[1], noofargs, arg);     
     if(g.tif_xsize>1) g.tif_xsize=1;
     if(g.tif_ysize>1) g.tif_ysize=1;
     if(g.tif_xsize<0) g.tif_xsize=0;
     if(g.tif_ysize<0) g.tif_ysize=0;
     in_read = 0;
  }
}


//--------------------------------------------------------------------------//
//  readcheck                                                               //
//--------------------------------------------------------------------------//
void readcheck(dialoginfo *a, int radio, int box, int boxbutton)
{
   box=box; radio=radio; boxbutton=boxbutton;
   int k;
   int count = a->boxcount[1];
   int status=OK;
   char **filenames;
   char *ptr; 
   strcpy(read_filename, a->answer[1][0]);
   g.read_raw = a->radioset[0]-1;
   g.want_quantize = a->radioset[1];
   g.tif_xoffset = atoi(a->answer[2][0]);
   g.tif_yoffset = atoi(a->answer[3][0]);
   g.tif_xsize  = strtod(a->answer[4][0], &ptr)/100.0;
   g.tif_ysize  = strtod(a->answer[5][0], &ptr)/100.0;
   if(g.tif_xsize>1) g.tif_xsize = 1;
   if(g.tif_ysize>1) g.tif_ysize = 1;
   if(g.tif_xsize<0) g.tif_xsize = 0;
   if(g.tif_ysize<0) g.tif_ysize = 0;
   if(a->boxset[6]) g.read_cmyk     = 1; else g.read_cmyk     = 0;
   if(a->boxset[7]) g.read_grayscale= 1; else g.read_grayscale= 0;
   if(a->boxset[8]) g.read_convert  = 1; else g.read_convert  = 0;
   if(a->boxset[9]) g.tif_positive  = 0; else g.tif_positive  = 1;
   if(a->boxset[10]) g.want_shell   = 1; else g.want_shell    = 0;
   if(a->boxset[11]) g.read_swap_bytes = 1; else g.read_swap_bytes = 0;

   //// 12 is a label
   g.want_decompression = a->boxset[13];   
   strcpy(g.decompression, a->answer[14][0]);
   strcpy(g.compression_ext, a->answer[15][0]);

   g.want_decryption = a->boxset[16];
   strcpy(g.decryption, a->answer[17][0]);
   strcpy(g.encryption_ext, a->answer[18][0]);

   g.read_split_frames = a->boxset[19];
   g.fix_cmyk = a->boxset[20];

   filenames = a->boxlist[1];  
   if(radio == -2 && filenames != NULL)   //// User clicked Ok or Enter
   {   for(k=0;k<count;k++)
       {   status = readimage(filenames[k], 0, NULL);
           status_error_message(status);
       }
   }
}


//--------------------------------------------------------------------------//
//  readfinish                                                              //
//--------------------------------------------------------------------------//
void readfinish(dialoginfo *a)
{
   if(!in_read) return;
   int k, count = a->boxcount[1];
   char **filenames = (char**)a->boxlist[1];  
   for(k=0;k<count;k++) XtFree(filenames[k]);
   if(count) delete[] filenames;
   g.getout=0;
   in_read = 0;
   selectimage(ci);
}


//--------------------------------------------------------------------------//
// readfiles                                                                //
// Reads one or more image files. Accepts wildcards.                        //
// Use this function instead of readimage() for loading images when there   //
// is a chance the filename may contain wildcard characters.                //
// Returns the error status.                                                //
//--------------------------------------------------------------------------//
int readfiles(char* filename, int noofargs, char **arg)
{
   if(memorylessthan(4096)){  message(g.nomemory,ERROR); return(0); } 
   int status=OK;

   // In Unix, wildcard expansion is done by the shell. So we only 
   // have to read the specified file. Set skipswitch to 0 to prevent
   // redrawing until the end.

   if(filename!=NULL) 
   {  g.read_skipswitch = 1;
      status = readimage(filename, noofargs, arg);
      g.read_skipswitch = 0;
   }   
   selectimage(ci);
   switchto(ci);
   g.want_switching=1;

   //// fix for the fix needed to counteract cinnamon and some other
   //// window managers which try to take over window focusing
   
   g.ulx = 0;
   g.uly = 0;
   g.lrx = z[ci].xsize;
   g.lry = z[ci].ysize;
   return(status);
}


//--------------------------------------------------------------------------//
// readimage                                                                //
// Determines whether image is a TIF (=1), PCX (=2), IMA (3), IMG (4), etc. //
// Then automatically calls the appropriate file reading routine.           //
// This function only loads a single file.                                  //
// Use readfiles(), which checks for wildcards, to load an image with a     //
// user-specified filename which could have wildcards.                      //
//--------------------------------------------------------------------------//
int readimage(char* filename, int noofargs, char **arg)
{   
   if(g.diagnose){ printf("Reading image %s\n",filename);fflush(stdout); }
   int type, status=OK, oci=ci, need_processing = 0;;
   char original_filename[FILENAMELENGTH];
   char tempstring[3*FILENAMELENGTH];
   char temp_filename[FILENAMELENGTH]="";
   char last_ext[64];
   char all_ext[64];
   char second_last_ext[64];
   char fname[1024]="";
   int compressed=0, encrypted=0;

   ////  Test if file is readable by trying to open it - a more thorough 
   ////  test than access(). Also check if it's a directory.
   FILE *fp;
   if(is_dir(filename)) 
   {   sprintf(tempstring, "%s\nis a directory", filename);
       message(tempstring, ERROR); 
       return ERROR; 
   }   
   if((fp=fopen(filename,"r")) == NULL)
   {   error_message(filename, errno);
       return(ERROR);
   }
   fclose(fp);  //// no longer reading the file
   g.invertbytes=0;

   g.read_signedpixels = 0;
   g.state=READING;
   printstatus(g.state);
   drawselectbox(OFF);
   setSVGApalette(1,0); // Switch to grayscale in case image is 8 bit & has no palette

   strcpy(read_filename, filename);
   strcpy(original_filename, filename);
   find_last_two_extensions(filename, last_ext, second_last_ext);
   get_extension(filename, all_ext);
   if(strstr(all_ext, g.compression_ext) || strstr(all_ext, g.encryption_ext))
       need_processing = 1;

   //// Assume encryption and/or compression are last two extensions
   //// Can be in either order
   //// These change filename
   //// Do everything on a temporary file, then delete it

   if(need_processing)
   {    
         create_temporary_file(temp_filename, filename);
         if(!strcmp(last_ext, g.encryption_ext)) { encrypted=1; decrypt(temp_filename);}
         if(!strcmp(last_ext, g.compression_ext)) { compressed=1; decompress(temp_filename);}
         if(!strcmp(second_last_ext, g.encryption_ext)){ encrypted=2; decrypt(temp_filename); }
         if(!strcmp(second_last_ext, g.compression_ext)) { compressed=2; decompress(temp_filename);}
         z[ci].encrypted = encrypted;
         z[ci].compressed = compressed;
         strcpy(fname, temp_filename);
   }else strcpy(fname, filename);

   imagefiletype(fname, identifier, compressed);  //// now we can find out image type
   switch(g.read_raw)
   {   case 1:  type = RAW; break;
       case 2:  type = ASCII; break;
       case 3:  type = RAW3D; break;
       default: type = imagefiletype(fname, identifier, compressed);
   }

   printstatus(READING);
   switch(type)       // status==OK means no errors
   {   case TIF:   status=readtiffile(fname, fname);     break;
       case PCX:   status=readpcxfile(fname);     break;
       case IMA:   status=readimafile(fname);     break;
       case IMM:
       case IMG:
       case RAW:
       case ASCII: 
                   noofargs = max(noofargs, 2);
                   strcpy(g.raw_filename, fname); // don't pass string, goes out of scope
                   strcpy(g.raw_filename2, filename);
                   status=readimgfile(type,noofargs,arg,compressed,identifier);
                   break;
       case RAW3D: status=read3Dfile(fname);      break;
       case MULTIFRAME: status=readmultiframefile(fname);      break;
       case GIF:   status=readgiffile(fname);     break;
       case GEM:
       case IMDS:  status=readgemfile(fname,type);break;
       case JPG:   status=readjpgfile(fname);     break;
       case BMP:   status=readbmpfile(fname);     break;
       case TARGA: status=readtgafile(fname);     break;
       case PICT:  status=readjpgfile(fname);     break;
       case XWD:   status=readxwdfile(fname);     break;
       case FITS:  status=readfitsfile(fname);    break;
       case PDS:   status=readpdsfile(fname);     break;
       case PDS2:  status=readpds2file(fname);    break;
       case PBM:    
       case PGM:    
       case PPM:  
       case PBMRAW: 
       case PGMRAW: 
       case PPMRAW: status=readpbmfile(fname,0); break;
       case PNG:    status=readpngfile(fname); break;
       case P2:     status=readp2file(fname); break;
       case P3:     status=readp3file(fname); break;
       case P4:     status=readp4file(fname); break;
       case BIORAD: status=readbioradfile(fname);break;  
       case TEXT:   status=readtextfile(fname); break;
       case SCN:    status=readscnfile(fname);break;  
   }   

   //// For encryption
   if(need_processing) delete_temporary_file(temp_filename);

   //// Only executed if image reading function does not call a dialog.
   //// Image reading function sets g.read_signedpixels if necessary.

   if(ci != oci) initialize_image(fname, filename, ci, type, compressed, 
                 g.read_signedpixels, identifier, status);

   printstatus(NORMAL);
   return OK;
}


//--------------------------------------------------------------------------//
// delete_temporary_file                                                    //
//--------------------------------------------------------------------------//
void delete_temporary_file(char *temp_filename)
{
   char tempstring[FILENAMELENGTH];
   int status;
   sprintf(tempstring, "rm %s", temp_filename);  
   status = system(tempstring);                      //// get rid of tempfile.png.gz.aes
   if(status != 0) { message("Error deleting temporary file, aborting"); return;}
}

//--------------------------------------------------------------------------//
// create_temporary_file                                                    //
//--------------------------------------------------------------------------//
void create_temporary_file(char *temp_filename, char *filename)
{
   char tempstring[2*FILENAMELENGTH];
   char all_ext[64];
   int status = 0;
   strcpy(temp_filename, "~temporary");
   get_extension(filename, all_ext);
   strcat(temp_filename, all_ext);
   //// Make a temporary copy, then decrypt it
   sprintf(tempstring, "cp %s %s", filename, temp_filename); 
   status = system(tempstring);                   
   if(status != 0) { message("Error creating tempfile, aborting"); return;}
}

//--------------------------------------------------------------------------//
// decompress removes .gz from filename and runs g.decompress.              //
// Returns 0 if success, error code if a problem occurs.                    //                                     
// Decompress program removes .gz extension automatically.                  //
// Deletes .gz file automatically but leaves uncompressed file              //
// so temp file needed. Changes filename[].                                 //
// filename[] must be a temporary file, not the real file, as it gets       //
// deleted.                                                                 //
//--------------------------------------------------------------------------//
int decompress(char *filename)
{  
   char tempstring[2*FILENAMELENGTH];
   int status = 0;
   sprintf(tempstring, "%s %s ", g.decompression, filename); 
   status = system(tempstring);                        //// deletes uncompressed file
   if(status != 0) { message("Error compressing file, aborting"); return ABORT;}
   remove_extension(filename, g.compression_ext);      //// now is tempfile.png
   return 0;
}


//--------------------------------------------------------------------------//
// decrypt - removes .aes from filename and runs g.deryption.               //
// Returns 0 if success, error code if a problem occurs.                    //                                     
// Decryption program removes .aes extension automatically                  //
// but doesn't delete the encrypted one.                                    //
// filename[] must be a temporary file, not the real file, as it gets       //
// deleted.                                                                 //
//--------------------------------------------------------------------------//
int decrypt(char *filename)
{
   char tempstring[3*FILENAMELENGTH];
   char program[128];
   int status = 0;
   //// Get aes password
   //// No password is set, or ask for password for each file
   if(!strlen(g.aespassword) || g.password_mode==0) 
   {   if(g.password_mode==0) sprintf(tempstring, "AES Password for\n%s:", filename);
       else strcpy(tempstring, "Password for AES files:");
       getstring(tempstring, g.aespassword, 1024, 78);
   }

   //// Chop off any params, then make sure it is executable.
   //// Note, everything must have full path or access doesn't find it.
   //// program must have full path name
   strcpy(program, g.decryption);                     //// copy of program w/o options
   set_array_position(program, ' ', 0);
   if(access(program, X_OK)) 
   {      sprintf(tempstring, "Cannot execute %s", program);
          message(tempstring);
          return ERROR;
   }

   strcpy(program, g.decryption);                     //// new copy of program w/options
   set_array_position(program, '#', 0);               //// slap that baby in
   sprintf(tempstring, "%s %s %s", program, g.aespassword, filename); 
   status = system(tempstring);                      //// now is tempfile.png.gz
   if(status != 0) { message("Error decrypting file, aborting"); return ABORT;}

   sprintf(tempstring, "rm %s",filename);            //// remove tempfile.png.gz.aes
   status = system(tempstring);                      //// now is tempfile.png.gz
   if(status != 0) { message("Error deleting temp file, aborting"); return ABORT;}
   remove_extension(filename, g.encryption_ext);     //// 
   return 0;
}


//--------------------------------------------------------------------------//
// initialize_image                                                         //
//--------------------------------------------------------------------------//
void initialize_image(char *fname, char *filename, int ino, int type, 
                      int compress, int signedpixels, char *identifier,
                      int status)
{
   int delay,i,i2,j,k,oci,byteswap = 0;
   uint color;

   char bfname[FILENAMELENGTH]; // basefile name
   char tempstring[FILENAMELENGTH];
   //// For image types whose bytes are not automatically swapped in Solaris
   if(g.ximage_byte_order != LSBFirst)
   {   if(g.bitsperpixel == 8)
       {   switch(type)
           {    case JPG:
                case BMP:
                case TARGA:
                case PNG:
                   byteswap = 1;
                   break;
           }
       }
   }
   if(compress)
   {   unlink(fname);
       z[ino].compressed = 1;
   }
   printstatus(g.state);
   if(!g.tif_positive) invertimage(ci); 

   //// Any other file formats with swapped bytes for grayscale, add them here
   if(type == PNG && z[ino].bpp > 8 && z[ino].colortype == GRAY) byteswap = 1;
   
   if(g.invertbytes) byteswap = 1 - byteswap;
   if(g.read_swap_bytes) byteswap = 1 - byteswap;
   if(type==ASCII) byteswap=0;
   if(byteswap) swap_image_bytes(ino);
                                // Must be done after bytes are swapped 
                                // Set in "read raw bytes"
   if(signedpixels && (type==RAW || type==ASCII)) 
      scale_signed_image(ino);
   if(ino>=0) setimagetitle(ino, strip_relative_path(filename));

   if(g.want_shell)
   {   
      strcpy(bfname, basefilename(filename));
      XtVaSetValues(z[ino].widget, XmNtitle, bfname, XmNiconName, bfname, (char *)0);
      XtVaSetValues(XtParent(z[ino].widget), XmNtitle, bfname, XmNiconName, bfname, (char *)0);
      XtVaSetValues(XtParent(XtParent(z[ino].widget)), XmNtitle, bfname, XmNiconName, bfname, (char *)0);
   }
   strcpy(z[ino].format_name, identifier);
   z[ino].format = type;

   if(status!=OK)
   {  sprintf(tempstring,"Error reading %s file",z[ino].format_name);
      message(tempstring,ERROR);
   }
   printstatus(g.state);
   if(g.read_grayscale) converttograyscale(ino);
                                 // Expand image if in 15-24 bpp modes
   if(g.read_convert) change_image_depth(ino,g.bitsperpixel,PERM);
   memcpy(z[ino].palette, g.palette, 768);
   memcpy(z[ino].opalette, g.palette, 768);
   memcpy(z[ino].spalette, g.palette, 768);
                                 // Check if image is actually grayscale

   if(g.autoundo) backupimage(ino,0); 
   if(g.read_skipswitch == 0) switchto(ino);
   setpalette(z[ino].palette);
   rebuild_display(ino);
   repair(ino);
                                 // Different pixel interact modes
   if(g.imode!=SET)
   {  for(j=z[ino].ypos;j<z[ino].ypos+z[ino].ysize;j++)
      { for(i=z[ino].xpos,i2=0;i<z[ino].xpos+z[ino].xsize;i++,
              i2+=g.off[z[ino].bpp])
        {  color = pixelat(z[ino].image[z[ino].cf][j-z[ino].ypos]+i2,z[ino].bpp);
                                 // Take color from the image
                                 // Interact it with topmost pixel from
                                 // another image or background but not
                                 // itself (skip ino).
           setpixelonimage(i,j,color,g.imode,z[ino].bpp,ino);
        }   
        if(keyhit()) if(getcharacter()==27) break;
      }
      if(g.read_skipswitch==0) switchto(ino);          
      repairimg(ino,0,0,z[ino].xsize,z[ino].ysize);             // Fix up display
   }
       
   if(g.diagnose)
   {  printf("done, ino %d x %d y %d bpp %d\n",ino,z[ino].xsize,z[ino].ysize,z[ino].bpp); }
   g.state=NORMAL;
   printstatus(g.state);

   if(g.read_split_frames)
   {   oci = ci;
       split_frames(z[ci].xpos+10, z[ci].ypos+50, ci, 0, z[ci].window_border, 
           g.create_cols, 1);
       for(k=oci+1; k<=oci+z[oci].frames; k++)
       {   z[k].split_frames = 1;
           z[k].split_frame_start = oci + 1;
           z[k].split_frame_end = oci + z[oci].frames;       
           z[k].oframe_count = z[oci].frames;
           if(g.autoundo) backupimage(k, 0); 
           switchto(oci+1);
       }
   }
   if(z[ino].animation && z[ino].fps) 
   {  delay = (int)(1000/z[ino].fps);
      movie(delay);
   }
   return;
}


//--------------------------------------------------------------------------//
// fix_negative_value - rescale signed pixel to unsigned range              //
//--------------------------------------------------------------------------//
uint fix_negative_value(int v, int bpp)
{
   switch(bpp)
   {  case 7:
      // FALLTHRU 
      // stupid comment needed to shut up the compiler
      case 8:  if(v & 0x80) return 255 - (v & 0x7f); 
      // FALLTHRU 
      case 10: return 1024 - v;
      case 12: return 4096 - v;
      // FALLTHRU
      case 15:
      case 16: if(v & 0x8000) return 32767 - (v & 0x7fff);
      // FALLTHRU
      case 24:
      case 32: if(v & 0x800000) return 16777215 - (v & 0x7fffff); 
      // FALLTHRU
      default: 
      return (uint)v;
   }
}


//--------------------------------------------------------------------------//
// scale_signed_image                                                       //
//--------------------------------------------------------------------------//
void scale_signed_image(int ino)
{
   int value, bpp, r, f, i, j;
   bpp = z[ino].bpp;
   r = g.off[bpp];
   for(f=0; f<z[ino].frames; f++)
   for(j=0; j<z[ino].ysize; j++)
   for(i=0; i<r*z[ino].xsize; i+=r)
   {  value = pixelat(z[ino].image[f][j]+i, bpp);
      value = fix_negative_value(value, z[ino].originalbpp);
      putpixelbytes(z[ino].image[f][j]+i,value,1,bpp,1);
   }   
   repair(ino);
}


//--------------------------------------------------------------------------//
// saveimage                                                                //
// Entry point for saving image on disk                                     //
//--------------------------------------------------------------------------//
int saveimage(int ino, int noofargs, char **arg)
{
  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return NOMEM; }
  static Dialog *dialog;
  static char filename[2*FILENAMELENGTH];   
  static int file_format=TIF;
  int compress;
  static int selection=0;
  static int ascii_selection=1;

  if(in_save) return BUSY;
  in_save = 1;

  int bpp,j,k; 
  combinedino = -1;
  if(ino>0 && strlen(z[ino].name)>0) 
  {    if(strchr(z[ino].name, '/'))                // If name has path
           strcpy(filename,z[ino].name);
       else                                        // If no path, add cwd
           sprintf(filename,"%s/%s",g.currentdir,z[ino].name);
       bpp = z[ino].bpp;
       strcpy(identifier, z[ino].format_name);
       file_format = z[ino].format;
  }else 
  {    sprintf(filename,"%s/Untitled",g.currentdir);
       bpp = g.bitsperpixel;

       strcpy(identifier, g.default_image_format_name);
       file_format = which_file_format(g.default_image_format_name);
  }

  noofformats = STDFORMATS;
  if(file_format > noofformats) file_format = g.default_image_format;
  if(ino>=0) compress = z[ino].compressed;

  if(noofargs)
  {    if(noofargs>=1) strcpy(filename, arg[1]);
       if(noofargs>=2) file_format = atoi(arg[2]);
       if(noofargs>=3) g.want_bpp = atoi(arg[3]);
       if(noofargs>=4) compress = atoi(arg[4]);
       save_image_file(filename, file_format, ino, 1);        
       in_save = 0;
       in_save_dialog = 0;
       blocked_while_saving = 0;
       g.block = max(0,g.block-1);
  }else
  {    dialog = new Dialog;
       in_save_dialog = dialog;
       if(dialog==NULL){ message(g.nomemory); in_save = 0;  in_save_dialog = 0; return NOMEM; }
       strcpy(dialog->title,"Save Image");
       strcpy(dialog->radio[0][0],"Image type");             
       strcpy(dialog->radio[0][1],"1 bpp Monochrome");    // want_color_type==0
       strcpy(dialog->radio[0][2],"8 bpp Gray Scale");    // want_color_type==1
       strcpy(dialog->radio[0][3],"8 bpp Indexed Color"); // want_color_type==2
       strcpy(dialog->radio[0][4],"15 bits/pixel");       // want_color_type==3
       strcpy(dialog->radio[0][5],"16 bits/pixel");       // want_color_type==4
       strcpy(dialog->radio[0][6],"24 bits/pixel");       // want_color_type==5
       strcpy(dialog->radio[0][7],"32 bits/pixel");       // want_color_type==6
       strcpy(dialog->radio[0][8],"48 bits/pixel");       // want_color_type==7

       strcpy(dialog->radio[1][0],"Treat data as");             
       strcpy(dialog->radio[1][1],"Color");
       strcpy(dialog->radio[1][2],"Gray scale");

       //// Default radio settings

       if(ino>=0 && z[ino].colortype==GRAY)
       {     dialog->radioset[1]=2;
             g.usegrayscale=1;
       }else
       {     dialog->radioset[1]=1;
             g.usegrayscale=0;
       }

       set_want_color_type(bpp);  //// set appropriate globals for file formats

       dialog->radioset[0] = g.want_color_type+1;
       dialog->noofradios=2;
       dialog->radiono[0]=9;
       dialog->radiono[1]=3;
       dialog->radiono[2]=5;
       dialog->radiono[3]=0;
       for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
   
       ////  Boxes  

       strcpy(dialog->boxes[0],"File Parameters");
       strcpy(dialog->boxes[1],"Filename");
       strcpy(dialog->boxes[2],"Format");
       strcpy(dialog->boxes[3],"Image no.");
       strcpy(dialog->boxes[4],"Extra Parameters (TIF)");
       strcpy(dialog->boxes[5],"Primary colors");
       strcpy(dialog->boxes[6],"Red bits/pixel");
       strcpy(dialog->boxes[7],"Green bits/pixel");
       strcpy(dialog->boxes[8],"Blue bits/pixel");
       strcpy(dialog->boxes[9],"Black bits/pixel");
       strcpy(dialog->boxes[10],"RGB->CMYK");
       strcpy(dialog->boxes[11],"JPEG quality");
       strcpy(dialog->boxes[12] ,"Post-processing");
       strcpy(dialog->boxes[13] ,"Compress file");
       strcpy(dialog->boxes[14] ,"Compress Command");
       strcpy(dialog->boxes[15] ,"Compress Extension");
       strcpy(dialog->boxes[16] ,"Encrypt file");
       strcpy(dialog->boxes[17] ,"Encrypt Command");
       strcpy(dialog->boxes[18] ,"Encrypt Extension");
       strcpy(dialog->boxes[19] ,"Misc. parameters");
       strcpy(dialog->boxes[20] ,"ASCII Format");

       //// Default box values

       strcpy(dialog->answer[1][0],filename);
       //// 2 is below
       sprintf(dialog->answer[3][0],"%d",ino);
       sprintf(dialog->answer[5][0],"%d",g.want_noofcolors);
       sprintf(dialog->answer[6][0],"%d",g.want_redbpp);
       sprintf(dialog->answer[7][0],"%d",g.want_greenbpp);
       sprintf(dialog->answer[8][0],"%d",g.want_bluebpp);
       sprintf(dialog->answer[9][0],"%d",g.want_blackbpp);
       dialog->boxset[10] = g.save_cmyk;
       sprintf(dialog->answer[11][0],"%d",g.jpeg_qfac);
       //// 12 is a label
       dialog->boxset[13] = compress;
       strcpy(dialog->answer[14][0], g.compression);
       strcpy(dialog->answer[15][0], g.compression_ext);

       dialog->boxset[16] = g.want_encryption;
       strcpy(dialog->answer[17][0], g.encryption);
       strcpy(dialog->answer[18][0], g.encryption_ext);
       //// 19 is a label
       sprintf(dialog->answer[20][0],"%d",g.want_bpp);

       //// List for file format selection box 2
       dialog->l[2]        = new listinfo;
       dialog->l[2]->title = new char[100];
       dialog->l[2]->title[0] = 0;
       dialog->l[2]->info  = new char*[noofformats];
       dialog->l[2]->count = noofformats;
       dialog->l[2]->wantfunction = 0;
       dialog->l[2]->f1    = null;
       dialog->l[2]->f2    = null;
       dialog->l[2]->f3    = null;
       dialog->l[2]->f4    = null; // dialog lists are deleted in dialogcancelcb
       dialog->l[2]->f5    = null;
       dialog->l[2]->f6    = null;
       dialog->l[2]->listnumber = 0;
       for(k=0; k<noofformats; k++)
       {    dialog->l[2]->info[k] = new char[100]; 
            dialog->l[2]->info[k][0] = 0;
       } 
       selection = format_selection(file_format);
       dialog->l[2]->selection = &selection;

       ////  Do not exceed 100 characters for the following menu items
       strcpy(dialog->l[2]->title, "File Format");
       for(k=0; k<STDFORMATS; k++)
           sprintf(dialog->l[2]->info[k], "%s %s", " ",which_file_identifier(k));

       //// List for ascii format params box 20
       dialog->l[20]        = new listinfo;
       dialog->l[20]->title = new char[100];
       dialog->l[20]->title[0] = 0;
       dialog->l[20]->info  = new char*[4];
       dialog->l[20]->count = 4;
       dialog->l[20]->wantfunction = 0;
       dialog->l[20]->f1    = null;
       dialog->l[20]->f2    = null;
       dialog->l[20]->f3    = null;
       dialog->l[20]->f4    = null; // dialog lists are deleted in dialogcancelcb
       dialog->l[20]->f5    = null;
       dialog->l[20]->f6    = null;
       dialog->l[20]->listnumber = 0;
       for(k=0; k<4; k++)
       {    dialog->l[20]->info[k] = new char[100]; 
            dialog->l[20]->info[k][0] = 0;
       } 
       switch(g.save_ascii_format)
       {    case INTEGER:  ascii_selection=0; break;
            case DOUBLE:   ascii_selection=1; break;
            case RGBVALUE: ascii_selection=2; break;
            case CALIBRATED_VALUE: ascii_selection=3; break;
            default: ascii_selection=0;
       }
       dialog->l[20]->selection = &ascii_selection;
       ////  Do not exceed 100 characters for the following menu items
       strcpy(dialog->l[20]->title, "ASCII Format");
       strcpy(dialog->l[20]->info[0]," Integer             ");
       strcpy(dialog->l[20]->info[1]," Density (0-1)       ");
       strcpy(dialog->l[20]->info[2]," RGB                 ");
       strcpy(dialog->l[20]->info[3]," Calibrated value    ");

       dialog->boxtype[0]=LABEL;        
       dialog->boxtype[1]=FILENAME;       
       dialog->boxtype[2]=NON_EDIT_LIST;
       dialog->boxtype[3]=INTCLICKBOX;
       dialog->boxtype[4]=LABEL;        
       dialog->boxtype[5]=INTCLICKBOX;
       dialog->boxtype[6]=INTCLICKBOX;
       dialog->boxtype[7]=INTCLICKBOX;
       dialog->boxtype[8]=INTCLICKBOX;  
       dialog->boxtype[9]=INTCLICKBOX;  
       dialog->boxtype[10]=TOGGLE;      
       dialog->boxtype[11]=INTCLICKBOX; 
       dialog->boxtype[12]=LABEL; 
       dialog->boxtype[13]=TOGGLE;      
       dialog->boxtype[14]=STRING;      
       dialog->boxtype[15]=STRING;
       dialog->boxtype[16]=TOGGLE;       
       dialog->boxtype[17]=STRING;      
       dialog->boxtype[18]=STRING;
       dialog->boxtype[19]=LABEL; 
       dialog->boxtype[20]=NON_EDIT_LIST;

       dialog->boxmin[2]=0;  dialog->boxmax[2]=noofformats;
       dialog->boxmin[3]=1;  dialog->boxmax[3]=g.image_count-1;
       dialog->boxmin[5]=1;  dialog->boxmax[5]=4;
       dialog->boxmin[6]=0;  dialog->boxmax[6]=32;
       dialog->boxmin[7]=0;  dialog->boxmax[7]=32;
       dialog->boxmin[8]=0;  dialog->boxmax[8]=32;
       dialog->boxmin[9]=0;  dialog->boxmax[9]=32;
       dialog->boxmin[11]=1; dialog->boxmax[11]=99;
       dialog->boxmin[17]=1; dialog->boxmax[17]=32;
       dialog->boxmin[20]=0; dialog->boxmax[20]=4;

       dialog->noofboxes=21;
       dialog->helptopic=3;
       dialog->want_changecicb = 1;  // if set to 1, it will change dialog box in f1
       dialog->f1 = savecheck;
       dialog->f2 = null;
       dialog->f3 = unblock_save;
       dialog->f4 = savefinish;
       dialog->f5 = save_change_image;
       dialog->f6 = null;
       dialog->f7 = null;
       dialog->f8 = null;
       dialog->f9 = null;
       dialog->width = 450;
       dialog->height = 0; // calculate automatically
       dialog->transient = 1;
       dialog->wantraise = 0;
       dialog->radiousexy = 0;
       dialog->boxusexy = 0;
       dialog->radioextrawidth = 0;
       strcpy(dialog->path,".");
       dialog->message[0] = 0;      
       dialog->busy = 0;
       dialogbox(dialog);
  }
  return OK;
}   

//--------------------------------------------------------------------------//
//  set_want_color_type                                                     //
//  sets globals needed by tiff and some other file formats                 //
//--------------------------------------------------------------------------//
void set_want_color_type(int bpp)
{
       switch(bpp)     
       {   case 8:  if(g.usegrayscale) g.want_color_type=1; 
                    else g.want_color_type=2; 
                    break;
           case 15: g.want_color_type=3; break;
           case 16: g.want_color_type=4; break;
           case 24: g.want_color_type=5; break;
           case 32: g.want_color_type=6; break;
           case 48: g.want_color_type=7; break;
       }
}

//--------------------------------------------------------------------------//
//  save_change_image                                                       //
//--------------------------------------------------------------------------//
void save_change_image(clickboxinfo *c)
{
    dialoginfo *a = (dialoginfo*)c->client_data;
    update_save_parameters(a, ci);
}


//--------------------------------------------------------------------------//
//  update_save_parameters                                                  //
//--------------------------------------------------------------------------//
void update_save_parameters(dialoginfo *a, int ino)
{
      char tempstring[128];
      static char filename[2*FILENAMELENGTH];
      int compress;
      sprintf(filename, "%s/%s", g.currentdir, basefilename(z[ino].name));
      strcpy(a->answer[1][0], filename);

      compress = z[ino].compressed;
      g.want_compression = compress;
      a->boxset[13] = compress;
      itoa(ino, a->answer[3][0], 10);

      a->radioset[1]=1; 
      set_want_color_type(z[ino].bpp);
      if(z[ino].colortype==GRAY){ a->radioset[1]=2; g.usegrayscale=1;} 
      else { a->radioset[0] = g.want_color_type+1; g.usegrayscale=0;}

      g.want_bpp = z[ino].bpp;
      
      set_widget_string(a->boxwidget[1][0], filename);
      set_widget_label(a->boxwidget[1][0], filename);
      itoa(ino, tempstring, 10);
      set_widget_label(a->boxwidget[3][0], tempstring);
      set_widget_label(a->boxwidget[2][0], z[ino].format_name);
      set_widget_toggle(a->boxwidget[13][0], compress);     
      set_want_color_type(z[ino].bpp);
      a->radioset[0] = g.want_color_type+1;

      set_widget_radio(a, 0, a->radioset[0]);
      set_widget_radio(a, 1, a->radioset[1]);
}


//--------------------------------------------------------------------------//
//  savecheck                                                               //
//--------------------------------------------------------------------------//
void savecheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  static int oci=-1, oxsize=-1, oysize=-1;
  static int oselectedimage = -1;
  static char filename[FILENAMELENGTH];

  if(!in_save){ message("Error",ERROR); return; }
  ////  If a new std. color type is added, add more entries to this table 
  ////  and inrease dimension of arrays if necessary   
  g.getout = 0;

  radio=radio; box=box; boxbutton=boxbutton;
  int file_format=TIF,ino,j,k,tif=0,jpeg=0, compress=0, xsize=0,ysize=0;
  int selection;
  //ino = ci;
  ino = atoi(a->answer[3][0]);

  selection = *(int*)a->l[2]->selection;
  int ascii_selection = *(int*)a->l[20]->selection;
  file_format = selection;
  strcpy(z[ino].format_name, which_file_identifier(file_format));
  strcpy(identifier, z[ino].format_name);

  if(ino>=0){ xsize = z[ino].xsize; ysize = z[ino].ysize; }

  //// update everything if they changed the image no.
  
  if(ino != oci || oselectedimage != g.selectedimage || xsize !=oxsize || ysize != oysize)
  {   
      oci = ino;
      oxsize = xsize;
      oysize = ysize;
      oselectedimage = g.selectedimage;
      update_save_parameters(a, ci);
  }
  compress = a->boxset[13];
  g.want_compression = compress;
  g.want_encryption = a->boxset[16];

  file_format = selection;
  if(file_format==TIF) tif=1;
  if(file_format==JPG) jpeg=1;

  for(k=0; k<3; k++)
  {   for(j=5; j<=10; j++)
      {  if(a->boxwidget[j][k])
         {   if(tif)  setsensitive(a->boxwidget[j][k],True);
             else     setsensitive(a->boxwidget[j][k],False);
      }  }
      if(a->boxwidget[11][k])
      {   if(jpeg)  setsensitive(a->boxwidget[11][k],True);
          else      setsensitive(a->boxwidget[11][k],False);
      }
  }
  if(file_format==ASCII)
  {   setsensitive(a->boxwidget[20][0],True);
      message("Warning: ASCII files are write-only");
      switch(ascii_selection)
      {    case 0: g.save_ascii_format = INTEGER; break;
           case 1: g.save_ascii_format = DOUBLE; break;
           case 2: g.save_ascii_format = RGBVALUE; break;
           case 3: g.save_ascii_format = CALIBRATED_VALUE; break;
      }
  }else   
      setsensitive(a->boxwidget[20][0],False);
  for(k=14; k<=15; k++)
      if(compress) setsensitive(a->boxwidget[k][0],True);
      else setsensitive(a->boxwidget[k][0],False);
  if(a->radioset[0] == 8)
    setsensitive(a->boxwidget[20][0],True);
  else
      setsensitive(a->boxwidget[20][0],False);

  //// Save user settings
  
  for(k=1; k<=8; k++) if(a->radioset[0]==k) g.want_color_type=k-1;
  if(a->radioset[1]==2) g.usegrayscale=1; else g.usegrayscale=0;

  g.jpeg_qfac = atoi(a->answer[11][0]);
  if(g.jpeg_qfac<1) g.jpeg_qfac=80;
  g.want_noofcolors= w_noofcolors[g.want_color_type];
  g.want_redbpp    = w_redbpp[g.want_color_type];
  g.want_greenbpp  = w_greenbpp[g.want_color_type];
  g.want_bluebpp   = w_bluebpp[g.want_color_type];
  g.want_blackbpp  = w_blackbpp[g.want_color_type];
  g.save_cmyk      = w_savecmyk[g.want_color_type];
  g.want_bpp       = w_bpp[g.want_color_type];
  set_want_color_type(g.want_bpp);

  if(strlen(a->answer[14][0])) strcpy(g.compression, a->answer[14][0]);
  if(strlen(a->answer[15][0])) strcpy(g.compression_ext, a->answer[15][0]);

  if(strlen(a->answer[17][0])) strcpy(g.encryption, a->answer[17][0]);
  if(strlen(a->answer[18][0])) strcpy(g.encryption_ext, a->answer[18][0]);

  if(g.usegrayscale) g.want_noofcolors = 1;

  //// If user clicked ok button in dialog, save image

  if(radio == -2 && in_save) 
  { 
     strcpy(filename, a->answer[1][0]);
     ino = atoi(a->answer[3][0]);
     ino = min(g.image_count-1, ino);
     if(!between(ino, 0, g.image_count-1)) message("Image not found");
     else save_image_file(filename, file_format, ino, 1);
  }
  return;
}


//--------------------------------------------------------------------------//
// format_selection                                                         //
//--------------------------------------------------------------------------//
int format_selection(int file_format)
{  
  int selection=0;
  selection = file_format;
  return selection;
}


//--------------------------------------------------------------------------//
// which_file_identifier                                                    //
//--------------------------------------------------------------------------//
char *which_file_identifier(int format)
{
   static char identifier[64];
   strcpy(identifier, "Unknown");
   //// only the first STDFORMATS are valid
   switch(format)
   {   case TIF: strcpy(identifier, "TIF"); break;
       case PCX: strcpy(identifier, "PCX"); break; 
       case IMG: strcpy(identifier, "IMG"); break;
       case GIF: strcpy(identifier, "GIF"); break;
       case IMM: strcpy(identifier, "IMM"); break;
       case GEM: strcpy(identifier, "GEM"); break;
       case IMDS: strcpy(identifier, "IMDS"); break;
       case RAW: strcpy(identifier, "RAW"); break;
       case PNG: strcpy(identifier, "PNG"); break;
       case P2:  strcpy(identifier, "P2"); break;
       case P3:  strcpy(identifier, "P3"); break;
       case P4:  strcpy(identifier, "P4"); break;
       case JPG: strcpy(identifier, "JPEG"); break;
       case TARGA: strcpy(identifier, "Targa"); break;
       case PICT: strcpy(identifier, "PICT"); break;
       case RLE: strcpy(identifier, "RLE"); break;
       case BMP: strcpy(identifier, "BMP"); break;
       case OS2: strcpy(identifier, "OS2"); break;
       case MAC: strcpy(identifier, "MAC"); break;
       case PC: strcpy(identifier, "PC"); break;
       case MVS: strcpy(identifier, "MVS"); break;
       case ASCII: strcpy(identifier, "ASCII"); break;
       case RAW3D: strcpy(identifier, "RAW3D"); break;
       case XWD: strcpy(identifier, "XWD"); break;
       case FITS: strcpy(identifier, "FITS"); break;
       case PDS: strcpy(identifier, "PDS"); break;
       case PDS2: strcpy(identifier, "PDS2"); break;
       case PGM: strcpy(identifier, "PGM"); break;
       case PPM: strcpy(identifier, "PPM"); break;
       case PBMRAW: strcpy(identifier, "PBMRAW"); break;
       case PGMRAW: strcpy(identifier, "PGMRAW"); break;
       case PPMRAW: strcpy(identifier, "PPMRAW"); break;
       case GIF89: strcpy(identifier, "GIF89"); break;
       case TEXT: strcpy(identifier, "TEXT"); break;
       case BIORAD: strcpy(identifier, "BIORAD"); break;
       case MULTIFRAME: strcpy(identifier, "MULTIFRAME"); break;
       case SCN: strcpy(identifier, "SCN"); break;
       case IMA: strcpy(identifier, "IMA"); break;
       default:strcpy(identifier, "unknown"); break;
   }
   return(identifier);
}


//--------------------------------------------------------------------------//
// which_file_format                                                        //
//--------------------------------------------------------------------------//
int which_file_format(char *identifier)
{
   int k=TIF;

   if(!strcasecmp(identifier, "TIF"))  k = TIF;
   if(!strcasecmp(identifier, "TIFF")) k = TIF;
   if(!strcasecmp(identifier, "PCX"))  k = PCX;
   if(!strcasecmp(identifier, "IMG"))  k = IMG;
   if(!strcasecmp(identifier, "JPEG")) k = JPG;
   if(!strcasecmp(identifier, "Targa")) k = TARGA;
   if(!strcasecmp(identifier, "TARGA")) k = TARGA;
   if(!strcasecmp(identifier, "TGA"))   k = TARGA;
   if(!strcasecmp(identifier, "BMP"))   k = BMP;
   if(!strcasecmp(identifier, "GIF"))   k = GIF89;
   if(!strcasecmp(identifier, "GIF87")) k = GIF;
   if(!strcasecmp(identifier, "GIF89")) k = GIF89;
   if(!strcasecmp(identifier, "ASC"))   k = ASCII;
   if(!strcasecmp(identifier, "ASCII")) k = ASCII;
   if(!strcasecmp(identifier, "RAW"))   k = RAW;
   if(!strcasecmp(identifier, "RAW3D")) k = RAW3D;
   if(!strcasecmp(identifier, "MULTIFRAME")) k = MULTIFRAME;
   if(!strcasecmp(identifier, "Multiframe")) k = MULTIFRAME;
   if(!strcasecmp(identifier, "FITS"))  k = FITS;
   if(!strcasecmp(identifier, "PNG"))   k = PNG;
   if(!strcasecmp(identifier, "P2"))    k = P2;
   if(!strcasecmp(identifier, "P3"))    k = P3;
   if(!strcasecmp(identifier, "P4"))    k = P4;
   if(!strcasecmp(identifier, "BIORAD")) k = BIORAD;
   if(!strcasecmp(identifier, "SCN")) k = SCN;
   return k;
}


//--------------------------------------------------------------------------//
//  unblock_save                                                            //
//--------------------------------------------------------------------------//
void unblock_save(dialoginfo *a)
{
   a=a;
   blocked_while_saving = 0;
   g.block = max(0, g.block-1);
}


//--------------------------------------------------------------------------//
//  savefinish                                                              //
//--------------------------------------------------------------------------//
void savefinish(dialoginfo *a)
{
   a=a;
   if(!in_save) return;
   if(between(combinedino, 1, g.image_count-1)) eraseimage(combinedino,0,0,0);
   combinedino = -1;
   hitnonstandard = 0;
   g.getout = 0;
   //// Unblock save_ok() which was waiting for user to save image before exiting
   blocked_while_saving = 0;
   in_save = 0; 
   in_save_dialog = 0;
}


//--------------------------------------------------------------------------//
//  save_all_images - save everything tagged w/o checking for overwrite     //
//  Don't compress or encrypt them.                                         //
//--------------------------------------------------------------------------//
int save_all_images(void)
{
  int kk, status=0, bad=0, count=0;
  int hit_format_check=0;
  int file_format;
  char filename[FILENAMELENGTH];
  char firstfilename[FILENAMELENGTH];
  char tempstring[3*FILENAMELENGTH];
  char tempstring2[2*FILENAMELENGTH];
  char aaa[16];
  char bbb[16];
  Widget www, scrollbar;

  in_save = 1;

  //// Bail out if file format is too complicated
  for(kk=1; kk<g.image_count; kk++)
  {       if(z[kk].tagged)
          {    file_format = z[kk].format;
               switch(file_format)
               {  case MULTIFRAME: bad=1; break;
                  default: count++;
               }
               if(z[kk].compressed) bad=1;
          }
  }
  if(bad){message("Unable to proceed\nsome images are in\n multiframe or compressed format"); return ERROR;}

  //// Get filename of first tagged one
  for(kk=1; kk<g.image_count; kk++)
  {    if(z[kk].tagged){ strcpy(firstfilename, z[kk].name); break;}
  }

  //// Bail out if none are tagged
  if(count==0)
  {    message("No tagged images to save");
       in_save = 0;
       return ABORT;
  }
  sprintf(tempstring,"Write %d tagged image(s)\nin directory\n%s\nStarting with %s?", 
       count, g.currentdir, firstfilename);

  if(message(tempstring, YESNOQUESTION)==YES)
  {  
      count = 0;
      progress_bar_open(www, scrollbar,(char*)"Progress");
      for(kk=1; kk<g.image_count; kk++)
      {  
          progress_bar_update(scrollbar, 100*kk/g.image_count);
          if(z[kk].tagged)
          {     strcpy(filename, z[kk].name);
                selectimage(kk);
                switchto(kk);
                file_format = z[kk].format;

                //// New images are always created in the default format

                if((file_format != g.default_image_format) && !hit_format_check)
                {
                   //// Compiler gets screwed up if you put a function call
                   //// in a sprintf. 
  
                   hit_format_check++;
                   strcpy(aaa, which_file_identifier(file_format));
                   strcpy(bbb, which_file_identifier(g.default_image_format));

                   sprintf(tempstring, "Warning: image %d is in %s format, but default format is %s.\n",
                         kk, aaa, bbb);
                   sprintf(tempstring2, "Click Yes to ignore or No to abort");
                   strcat(tempstring, tempstring2);
                   if(message(tempstring, YESNOQUESTION)!=YES)
                   {     message("File transfer canceled");
                         return CANCEL;
                   }
                }

                //// We have to set numerous globals here. Some file formats require  
                //// many details about bpp. Most are saved in globals
                //// The w_... are defined at top of xmtnimage3.cc
                //// This next part should go in a function

                set_want_color_type(z[kk].bpp);
                g.want_bpp = z[kk].bpp;
                g.want_noofcolors= w_noofcolors[g.want_color_type];
                g.want_redbpp    = w_redbpp[g.want_color_type];
                g.want_greenbpp  = w_greenbpp[g.want_color_type];
                g.want_bluebpp   = w_bluebpp[g.want_color_type];
                g.want_blackbpp  = w_blackbpp[g.want_color_type];
                g.save_cmyk      = w_savecmyk[g.want_color_type];
                g.want_bpp       = w_bpp[g.want_color_type];

                status = save_in_file_format(filename, file_format);
                if(status != OK)
                {
                    sprintf(tempstring, "Error saving image #%d\n%s\nOperation aborted", kk, z[kk].name);
                    message(tempstring, ERROR);
                    in_save = 0;
                    in_save_dialog = 0;
                    return status;
                }
                count++;
          }
      }
      progress_bar_close(www);
      sprintf(tempstring, "%d images saved", count);
      message(tempstring);
  }
  in_save = 0;
  in_save_dialog = 0;
  return status;
}

//--------------------------------------------------------------------------//
//  save_in_file_format                                                     //
//--------------------------------------------------------------------------//
int save_in_file_format(char *filename, int file_format)
{
   int status;
   printf("Saving image #%d %s\n",ci,filename);
   switch(file_format)
   {  
      case TIF: status = writetiffile(filename); break;
      case PCX: status = writepcxfile(filename); break;
      case IMG: status = writeimgfile(filename); break;
      case GIF: status = writegiffile(filename); break;
      case GIF89: status = writegif89file(filename); break;
      case JPG:   status = writejpeg(filename); break;
      case BMP:   status = writebmpfile(filename, BMP); break;
      case TARGA: status = writetgafile(filename); break;
      case ASCII: status = writeasciifile(filename); break;
      case RAW:   status = writerawfile(filename); break;
      case RAW3D: status = write3Dfile(filename); break;
      case FITS:  break;
      case PNG:   status = writepngfile(filename); break;
      case P2:    status = writep2file(filename); break;
      case P3:    status = writep3file(filename); break;
      case P4:    status = writep4file(filename); break;
      case MULTIFRAME: status = writemultiframefile(filename); break;
      default: message("Unsupported file format"); status = -1;         
   }
   return status;
}


//--------------------------------------------------------------------------//
//  save_image_file - called only by savecheck, don't call directly         //
//  if overwrite_check is set, ask about overwriting                       //
//--------------------------------------------------------------------------//
void save_image_file(char *filename, int file_format, int ino, int overwrite_check)
{
  int status=OK;
  static int param[10];
  static char actual_filename[FILENAMELENGTH];   
  static PromptStruct ps;
  char tempstring[1024];  

  if(ci<=0)
  {   sprintf(tempstring, "No image selected\nsave background?");
      if(message(tempstring, YESNOQUESTION) != YES) g.getout=1;
  }
  if(g.getout){  message("File transfer aborted", WARNING); g.getout=0; return; }

  ////  Check for consistency with specified file format
  int type = file_format;
  if(overwrite_check)
  {   switch(check_save_parameters(filename, ino, type))
      {   case OK: break;
          case ABORT: message("File transfer aborted", WARNING);  return;
          default: return;
      }
  }

  if(ino>=0) z[ino].compressed = g.want_compression;
  if(ino==-1) status=NOIMAGES;      // can't save, no images present
  if(g.getout){ g.getout=0; status=ABORT; }
  remove_extension(filename, g.encryption_ext);  // compress comes before encrypt
  if(status==OK)
  {    if(ino>=0 && z[ino].fftstate!=0) {  setSVGApalette(1,0); }
       strcpy(actual_filename, filename);
       param[0] = file_format;
       param[1] = 1;
       param[2] = ino;
       param[3] = 0;

       ps.filename = actual_filename; // filename to check for overwrite
       ps.text = filename;            // filename to use
       ps.f1 = save_image_file_part2;
       ps.f2 = null;
       ps.data = param;
       if(overwrite_check)
           check_overwrite_file(&ps);
       else
           prompt_yes(0, &ps, NULL);
  }else message("Bad parameters");
}  
      

//--------------------------------------------------------------------------//
//  save_image_file_part2                                                   //
//--------------------------------------------------------------------------//
void save_image_file_part2(PromptStruct *ps)
{
  int *param = (int*)ps->data;
  int file_format = param[0];
 
  //// Formats that require a second dialog box go here
  //// and call save_image_file_part3 when finished.
  //// All others go straight to save_image_file_part3.
 
  switch(file_format)
  {  
      case FITS: setupfitsfile(ps); break;
      default:  save_image_file_part3(ps);
  }      
}


//--------------------------------------------------------------------------//
//  save_image_file_part3                                                   //
//--------------------------------------------------------------------------//
void save_image_file_part3(PromptStruct *ps)
{
  char *filename  = ps->text;           // filename w/o .gz extension
  int *param = (int*)ps->data;
  char program[1024];
  int file_format = param[0];
  int ino         = param[2];
  int compress    = param[3];
  char tempstring[3*FILENAMELENGTH];   
  char tempstring2[FILENAMELENGTH];   
  char original_filename[FILENAMELENGTH];   
  char ext[64]="";
  int evenbits, savedxsize, savedysize, status=0;
  uint size;
  
  if(g.getout || !in_save){ message("Aborted"); return; }
  switchto(ino);      
  printstatus(WRITING);

  g.busy=1;

  //// Encrypt
  remove_extension(filename, g.encryption_ext); 
  remove_extension(filename, g.compression_ext); 
  get_extension(filename, ext);
  strcpy(original_filename, filename);
  if(g.want_encryption)
  {
         //// Get aes password
         //// No password is set, or ask for password for each file
         if(!strlen(g.aespassword) || g.password_mode==0) 
         {   if(g.password_mode==0) sprintf(tempstring, "AES Password for\n%s:", filename);
             else strcpy(tempstring, "Password for AES files:");
             getstring(tempstring, g.aespassword, 1024, 78);
         }

         //// Chop off any params, then make sure it is executable.
         //// Note, everything must have full path or access doesn't find it.
         sprintf(filename,"%s/%s%s", g.currentdir, "~tempfile", ext);
         strcpy(program, g.encryption);
         set_array_position(program, ' ', 0);
         if(access(program, X_OK)) 
         {      sprintf(tempstring, "Cannot execute %s", program);
                message(tempstring);
                return;
         }
  }

  //// FITS file formats have already been written in part2
  save_in_file_format(filename, file_format);
  if(status != 0) { message("Error saving file, aborting"); return;}
  g.busy=0;
  printstatus(NORMAL);

  if(compress)                                         //// should be same as when read
  {     sprintf(tempstring, "%s %s ", g.compression, filename); 
        status = system(tempstring);                   //// deletes uncompressed file
        if(status != 0) { message("Error compressing file, aborting"); return;}
        strcat(filename, g.compression_ext);           //// now is tempfile.png.gz            
        strcat(original_filename, g.compression_ext);  //// now is original.png.gz            
  }
  if(g.want_encryption)                                //// optional
  {     strcpy(program, g.encryption);
        set_array_position(program, '#', 0);
        sprintf(tempstring, "%s %s %s", program, g.aespassword, filename); 
        status = system(tempstring);                   //// now is tempfile.png.gz.aes
        if(status != 0) { message("Error encrypting file, aborting"); return;}
        sprintf(tempstring, "rm %s", filename);  
        status = system(tempstring);                   //// get rid of tempfile.png.gz
        if(status != 0) { message("Error deleting temporary file, aborting"); return;}
        strcat(original_filename, g.encryption_ext); 
        strcat(filename, g.encryption_ext); 
        sprintf(tempstring, "mv %s %s", filename, original_filename); 
        status = system(tempstring);                   //// now is original.png.gz.aes   
        if(status != 0) { message("Error renaming file, aborting"); return;}
        strcpy(filename, original_filename);         
        strcpy(z[ci].name, original_filename);         
  }
  size = filesize(filename);                           //// Check to ensure not an empty file
  if(size == 0) { message("Error handling file, aborting"); return; }
 

  if(status==OK && !access(filename, F_OK))           //// access = 0 if file exists
  {   savedxsize = z[ci].xsize;
      savedysize = z[ci].ysize;
      evenbits = ((savedxsize*g.want_bpp+7)/8)*8; 
      while(savedxsize*g.want_bpp < evenbits) savedxsize++;
      sprintf(tempstring,"%d x %d pixel region saved\nas %s format file\nin %s\n(size=%d bytes)",
          savedxsize, savedysize, 
          z[ino].format_name,
          filename, size);
      if(g.want_encryption)
      {    sprintf(tempstring2, "\nEncrypted size %d", filesize(filename));
           strcat(tempstring, tempstring2);
      }
      message(tempstring);
      if(ino>=0) z[ino].touched=0;
  }else 
  {   if(status==OK) 
      {    sprintf(tempstring, "File error: image may not have been saved\nError code:%d", errno);
           message(tempstring, ERROR);
      }else status_error_message(status);
  }   

  ////  Update image statistics

  if(status==OK && ino>=0)
  {   if(g.usegrayscale) z[ino].colortype=GRAY;
      if(ino>=0) setimagetitle(ino, filename);
      strcpy(z[ino].format_name, identifier);
      z[ino].format = file_format;
  }   
  return;

}

//--------------------------------------------------------------------------//
//  check_save_parameters                                                   //
//--------------------------------------------------------------------------//
int check_save_parameters(char *filename, int ino, int type)
{
  int status=OK, tot_colors, gota0=0, standard=1; 
  int tot_bpp = g.want_redbpp+g.want_greenbpp+g.want_bluebpp+g.want_blackbpp;
  char msg[FILENAMELENGTH] = "";  
  char extension[FILENAMELENGTH]="X";
  char extension_upper[FILENAMELENGTH]="X";
  char tempstring[FILENAMELENGTH];   
  char *ptr;
  switch(type)
  {   case TIF:
         tot_colors=0;
         {  if(g.want_blackbpp)tot_colors++;
            if(g.want_bluebpp) tot_colors++; else if(tot_colors)gota0=1;
            if(g.want_greenbpp)tot_colors++; else if(tot_colors)gota0=1;
            if(g.want_redbpp)  tot_colors++; else if(tot_colors)gota0=1;
         }
         if(g.want_noofcolors!=tot_colors && g.want_color_type>2 &&  g.want_noofcolors>1)
         {  strcpy(msg,"Number of colors doesn't add up.\nChange the bits/pixel on a color\nor number of primary colors");
            status=BADPARAMETERS;
         }                 
         if(g.want_noofcolors>1 && g.want_color_type==8 && g.want_bpp!=tot_bpp)
         {  strcpy(msg,"RGB bits/pixel don't add up to total.\nChange total bits/pixel or number of colors");
            status=BADPARAMETERS;
         }
         if(g.want_noofcolors>1 && g.usegrayscale && g.want_color_type==8)
         {  strcpy(msg,"You must have only 1 primary color\nto save data as grayscale");
            status=BADPARAMETERS;
         }
         if(g.want_noofcolors==1 && !g.usegrayscale && g.want_color_type==8)
         {  strcpy(msg,"You must have more than 1 primary color\nto save data as color");
            status=BADPARAMETERS;
         }
         if(g.save_cmyk && (g.want_blackbpp==0 || g.want_color_type<6) && g.want_color_type>2) 
         {  strcpy(msg,"Not enough bits for CMYK");
            status=BADPARAMETERS;
         }
         if(g.want_blackbpp && !g.save_cmyk)
         {  strcpy(msg,"You must select CMYK to use black bits");
            status=BADPARAMETERS;
         }
         if(g.want_noofcolors<4 && g.save_cmyk)
         {  strcpy(msg,"Must have 4 colors for CMYK"); 
            status=BADPARAMETERS;
         }
         if(g.usegrayscale && g.want_bpp>24)
         {  strcpy(msg,"Grayscale TIFF files cannot exceed 24 bits/pixel"); 
            status=BADPARAMETERS;
         }
         if(!hitnonstandard)
         {  standard = 1;                
            if(g.want_color_type==8)
            {  if(g.want_redbpp%8  || g.want_greenbpp%8) standard=0;
               if(g.want_bluebpp%8 || g.want_blackbpp%8) standard=0;
               if(g.want_bpp>8     && g.want_noofcolors<3) standard=0;
               if(g.want_bpp<=8    && g.want_noofcolors>1) standard=0;
               if(g.want_bpp<8     && g.want_bpp>1)        standard=0;
               if(g.want_bpp>24    && g.want_noofcolors<4) standard=0;
               if(g.want_bpp<=24   && g.want_noofcolors==4)standard=0;
               if(g.want_bpp==16   && g.want_redbpp==5 && g.want_greenbpp==6 && g.want_bluebpp==5) standard=1;
               if(g.want_bpp==15   && g.want_redbpp==5 && g.want_greenbpp==5 && g.want_bluebpp==5) standard=1;
               if(gota0) standard=0;                    
           }
           if(standard==0)   
           {  strcpy(msg,"Non-standard image format\nIs this ok?");
              if(message(msg,YESNOQUESTION)!=YES) status=ABORT;
              hitnonstandard = 1;
           }   
         }
         break;
      case GIF:  
      case GIF89:if(g.want_bpp!=8)
                 {  strcpy(msg,"GIF files must be 8 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 break;
      case PCX:  if(g.want_bpp!=8)
                 {  strcpy(msg,"PCX files must be 8 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 break;
      case IMG:  if(g.want_bpp!=1 && g.want_bpp!=8)
                 {  strcpy(msg,"IMG files must be 1 or 8 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 break;
      case RAW:
      case RAW3D:if(g.want_bpp!=8 && g.want_bpp!=15 && g.want_bpp!=16 && 
                    g.want_bpp!=24 && g.want_bpp!=32)
                 {  strcpy(msg,"Raw files must be 8,15,16,24,or 32 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 break;
      case FITS:     
                 if(g.want_bpp==16 && !g.usegrayscale)
                 {  strcpy(msg,"Color FITS files cannot be 16 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 if(g.want_color_type==2)  // indexed
                 {  strcpy(msg,"Indexed color FITS files not supported\nConvert to color first");
                    status=BADPARAMETERS;
                 }
                 break;
      case TARGA:if(g.want_bpp!=8 && g.want_bpp!=15 && g.want_bpp!=24 && g.want_bpp!=32)
                 {  strcpy(msg,"Targa files must be 8,15,24, \nor 32 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 strcpy(tempstring, filename);
                 ptr = strchr(tempstring,'.');
                 if(ptr!=NULL) strcpy(extension, ptr);  
                 strcpy(extension_upper, extension);
                 strupr(extension_upper);
                 if(strcmp(extension_upper,".TGA")) //TGA can only be id'd by ext'n
                 {  strcpy(msg,"Extension must be TGA for Targa format");
                    status=BADPARAMETERS;
                 }
                 break;
      case BMP:  if(g.want_bpp!=4 && g.want_bpp!=8 && g.want_bpp!=24)
                 {  strcpy(msg,"BMP files must be 4,8,or 24 bits/pixel");
                    status=BADPARAMETERS;
                 }
                 if(g.want_color_type==1)
                 {  strcpy(msg,"8 bit BMP files must have a palette");
                    status=BADPARAMETERS;
                 }
                 break;
      case JPG:  if(g.want_bpp != 24)
                    strcpy(msg,"Selected bit/pixel values\nwill be ignored");
                 if(g.usegrayscale)
                 {  strcpy(msg,"Image must be converted to color first");
                    status=BADPARAMETERS;
                 }
                 break;
      case PNG:  if(g.usegrayscale && g.want_bpp>16)
                 {  strcpy(msg,"Grayscale PNG files cannot exceed 16 bits/pixel"); 
                    status=BADPARAMETERS;
                 }
                 break;
      case P2:  if(g.want_bpp <16)
                {   status=BADPARAMETERS;
                    sprintf(msg,"Cannot save %d bit images in p2 format",g.want_bpp);
                }
                break;
      case P3:  if(g.want_bpp <16)
                {   status=BADPARAMETERS;
                    sprintf(msg,"Cannot save %d bit images in p3 format",g.want_bpp);
                }
                break;
      case P4:  if(g.want_bpp <16)
                {   status=BADPARAMETERS;
                    sprintf(msg,"Cannot save %d bit images in p4 format",g.want_bpp);
                }
                break;
      default:  break;           
  }
  if(filename[0]==0)
  {     strcpy(msg,"You must specify a filename");
        status=BADPARAMETERS;
  }
  if(ino>=0 && z[ino].bpp==48 && ((type != TIF && type != FITS)))
  {     strcpy(msg,"Must use TIFF or FITS format for 48-bit images");
        status=BADPARAMETERS;
  }
  if(status!=OK && strlen(msg)) message(msg, ERROR);

  //// Error messages requiring a response
  if( ino>=0 && g.want_color_type<=2 && z[ino].bpp>8 )
  {   strcpy(msg,"Color information will be lost!");
      strcat(msg,"\nYou should convert image to 8 bits/pixel first");
      strcat(msg,"\nOk to discard color information?");
      if(message(msg, YESNOQUESTION) != YES) status=ABORT;
  }
  if(z[ino].frames>1) 
  {   switch(type)
      {   case GIF89:
          case MULTIFRAME:
          case RAW3D: break;
          default:
             strcpy(msg,"Only current frame will be saved\nIs this ok?");
             if(message(msg, YESNOQUESTION) != YES) status=ABORT;
      }
  }
  if(ci==0) 
      if(message("Saving background", YESNOQUESTION) != YES) status=ABORT;
  return status;
}


//--------------------------------------------------------------------------//
// packbits                                                                 //
// Packs n bits into a block. Returns no.of bytes obtained.                 //
// Output block is the input bits packed into unsigned characters. This     //
//   means if n is a multiple of 8, the data are returned unchanged.        //
//   If not, the high-order bits from n to the next higher multiple of 8    //
//   are thrown away. The input data should                                 //
//   be right-shifted first if the high-order bits are to be saved.         //
//                                                                          //
// inpos     = Current position in array `inblock' (wraps to 0 if all the   //
//             bytes are used, i.e. if blockpos exceeds blocksize.          //
// outpos    = Current position in array `outblock'.                        //
// blocksize = Size of input array `inblock'                                //
// inblock   = char array of the data to be packed                          //
// outblock  = packed data which is returned                                //
//                                                                          //
// Bits are packed in the following format.                                 //
//                                                                          //
// Example for n=15:                                                        //
//     Input            7   6   5   4   3   2   1   0                       //
//     ------         --------------------------------                      //
//     Byte 1          a15 a14 a13 a12 a11 a10 a09 a08                      //
//     Byte 2          a07 a06 a05 a04 a03 a02 a01 a00                      //
//     Byte 3          b15 b14 b13 b12 b11 b10 b09 b08                      //
//     Byte 4          b07 b06 b05 b04 b03 b02 b01 b00                      //
//     ...                                                                  //
//     Output           7   6   5   4   3   2   1   0   Note: bit 15 is     //
//     ------         --------------------------------  thrown away.        //
//     Byte 1          a14 a13 a12 a11 a10 a09 a08 a07                      //
//     Byte 2          a06 a05 a04 a03 a02 a01 a00 b14                      //
//     Byte 3          b13 b12 b11 b10 b09 b08 b07 b06                      //
//     Byte 4          b05 b04 b03 b02 b01 b00 c14 c13                      //
//     ...                                                                  //
//                                                                          //
// Example for n=5:                                                         //
//     Input            7   6   5   4   3   2   1   0                       //
//     ------         --------------------------------                      //
//     Byte 1           a7  a6  a5  a4  a3  a2  a1  a0                      //
//     Byte 2           b7  b6  b5  b4  b3  b2  b1  b0                      //
//     Byte 3           c7  c6  c5  c4  c3  c2  c1  c0                      //
//     Byte 4           d7  d6  d5  d4  d3  d2  d1  d0                      //
//     ...                                                                  //
//     Output           7   6   5   4   3   2   1   0   Note: bits 7-4 are  //
//     ------         --------------------------------  thrown away. The    //
//     Byte 1           a4  a3  a2  a1  a0  b4  b3  b2  rest are packed     //
//     Byte 2           b1  b0  c4  c3  c2  c1  c0  d4  together.           //
//     Byte 3           d3  d2  d1  d0  e4  e3  e2  e1                      //
//     Byte 4           e0  f4  f3  f2  f1  f0  g4  g3                      //
//     ...                                                                  //
//--------------------------------------------------------------------------//
int packbits(int n,uchar* inblock,uchar* outblock, int blocksize)
{
   int k;
   int inpos = 0;   // input buffer byte position
   int outpos = 0;  // output buffer byte position
   int bitstoget=8; // no.of bits it is possible to get
   int need = 8;    // output bit position 8-0
   int total=0;     // input bit position 0-8
   int unusedbits;  // no.of bits to throw away
   int roundbytes;  // bytes in input for each n (n rounded up to next byte)
   int a;

   getbits(0,inblock,0,RESET);  //reset bits at each scan line

   roundbytes = (n+7) >> 3;
   for(k=0;k<blocksize;k++) outblock[k]=0;
   getbits(0,inblock,blocksize,RESET);
   unusedbits = (8 - (n%8))%8;  // 0,7,6,5,4,3,2,1,0,...
   getbits(unusedbits,inblock,blocksize,TIF);

   do
   {  bitstoget = min(need,n-total);
      a = getbits(bitstoget,inblock,blocksize,TIF) << (need-bitstoget);
      outblock[outpos] += a;
      total += bitstoget;
      need -= bitstoget;
      if(need==0) { outpos++; need=8; }
      if(total==n) 
      {  getbits(unusedbits,inblock,blocksize,TIF);
         total=0;
         inpos += roundbytes;
      }
   }while(inpos<blocksize);   
   return(outpos);
}


//--------------------------------------------------------------------------//
// imagefiletype                                                            //
// Returns integer TIF or whatever or -1 if unknown image file type.        //
// Returns string format identifier (must have enough for 64 bytes).        //
//--------------------------------------------------------------------------//
int imagefiletype(char *filename, char *identifier, int &compressed)
{  
   if(memorylessthan(16384)){ message(g.nomemory,ERROR); return(0); }
   FILE *fp;
   uchar header[1034], oheader[1034];
   char tempstring[1024], fitsheader[20], rawheader[20], newfilename[FILENAMELENGTH];
   char gifid[50], pbmid[5], extension[FILENAMELENGTH]="X";
   char extension_upper[FILENAMELENGTH]="X";
   int dot, k, h, ocompressed, pos, elength;
   short uint word1=0, word2=0, word3=0, word4=0;
   short int pcx_id=0, xsize, ysize;
   uchar pcx_version=0;
   long int fsize;
   
   compressed = 0;
   elength = (int)strlen(g.compression_ext);
   if(elength) pos = (int)strlen(filename) - elength; else pos=0;
   if(pos>=0 && strcmp(filename + pos, g.compression_ext)==SAME) compressed = 1;

   //// This will change 'compressed' if g.decompression is not found

   ocompressed = compressed;
   if((fp=open_file(filename, "r"))==NULL) return 0;
   compressed = ocompressed;

   if(compressed){ strncpy(newfilename, filename, pos); newfilename[pos]=0; }
   else strcpy(newfilename, filename);
   
   fread(header,1024,1,fp);
   strncpy(identifier, (char*)header, 63);
   identifier[63] = 0;
   memcpy(oheader, header, 128);
   close_file(fp);
 
   g.tif_skipbytes=0;

   //-------------Extract extension from filename------------//
   strcpy(tempstring, newfilename);
   dot = 0;
   for(k=0;k<(int)strlen(tempstring);k++) if(tempstring[k]=='.') dot=k;
   if(dot) strcpy(extension, tempstring+dot+1);

   //-------------Identify format----------------------------//
   for(h=0;h<=1;h++)
   { 
      word1 = header[0] + (header[1]<<8);
      word2 = header[2] + (header[3]<<8);
      word3 = header[4] + (header[5]<<8);
      word4 = header[6] + (header[7]<<8);
      pcx_id = word1 & 0xff;
      pcx_version = word1 & 0xff;
      word1 &= 0xffff;
      word2 &= 0xffff;
      strncpy(gifid, (char*)header, 5);
      gifid[5] = 0;
      pbmid[0]=header[0]; pbmid[1]=header[1]; pbmid[2]=0;
      override = 0;

      strncpy(fitsheader, (char*)header, 6);
      fitsheader[6] = 0;
      strncpy(rawheader, (char*)header, 10);
      rawheader[10] = 0;
      strcpy(extension_upper, extension);
      strupr(extension_upper);

      fitsheader[6]=0;
      //// put extension identifiers first
      if(!strcmp(extension_upper,"P2"))  
      {    strcpy(identifier,"P2");  return(P2); }
      else if(!strcmp(extension_upper,"P3"))  
      {    strcpy(identifier,"P3");  return(P3); }
      else if(!strcmp(extension_upper,"P4"))  
      {    strcpy(identifier,"P4");  return(P4); }
      else if(!strcmp(extension,"ascii"))  
      {    strcpy(identifier,"ASCII");  return(ASCII); }
      else if(!strcmp(extension,"txt"))  
      {    strcpy(identifier,"TEXT");  return(TEXT); }
      else if(!strcmp(extension,"scn"))  
      {    strcpy(identifier,"SCN");  return(SCN); }
      else if(!strcmp(extension,"SCN"))  
      {    strcpy(identifier,"SCN");  return(SCN); }
      else if(!strcmp(extension_upper,"TGA"))   // TGA can only be ident'd by extension
      {    strcpy(identifier,"Targa"); return(TARGA); }

      else if((word2==42)&&(word1==0x4949))
      {    strcpy(identifier,"TIF");  return(TIF); }            // Intel TIF
      else if((word2==0x2A00)&&(word1==0x4d4d))
      {    strcpy(identifier,"TIF");  return(TIF); }            // Macintosh TIF
      else if(pcx_id==0x0A)
      {    strcpy(identifier,"PCX");  return(PCX); }
      else if(!strcmp(gifid,"GIF87"))
      {    strcpy(identifier,"GIF87");  return(GIF); }       
      else if(!strcmp(gifid,"GIF89"))
      {    strcpy(identifier,"GIF89");  return(GIF); }       
      else if((word1==0x0100)&&((word2==0x0800)||(word2==0x0900)))// Regular GEM IMG
      {    strcpy(identifier,"GEM");  return(GEM); }
      else if((word1==0x0070)&&(word2==0x0191))    // IMDS = MVS mainframe GEM
      {    strcpy(identifier,"IMDS"); return(IMDS); }
      else if(((word1)==0xD8FF)&&((word2&0xFF)==0xFF))
      {    strcpy(identifier,"JPEG"); return(JPG); }
      else if((gifid[0]=='B')&&(gifid[1]=='M'))
      {    strcpy(identifier,"BMP");  return(BMP); }
      else if(word1==0x0000 && word3==0x0000 && word4==0x0700)
      {    strcpy(identifier,"XWD screen dump"); return(XWD); }
      else if(!strcmp(fitsheader,"SIMPLE"))  
      {    strcpy(identifier,"FITS"); return(FITS); }
      else if(!strcmp(rawheader,"Multiframe"))  
      {    strcpy(identifier,"MULTIFRAME"); return(MULTIFRAME); }
      else if(strstr((char*)header,"FILE_TYPE"))  
      {    strcpy(identifier,"PDS");  return(PDS); }
      else if(strstr((char*)header,"SFDU_LABEL"))  
      {    strcpy(identifier,"PDS2"); return(PDS2); }
      else if(!strcmp(pbmid,"P1"))  
      {    strcpy(identifier,"PBM");  return(PBM); }
      else if(!strcmp(pbmid,"P2"))  
      {    strcpy(identifier,"PGM");  return(PGM); }
      else if(!strcmp(pbmid,"P3"))  
      {    strcpy(identifier,"PPM");  return(PPM); }
      else if(!strcmp(pbmid,"P4"))  
      {    strcpy(identifier,"PBMRAW");  return(PBMRAW); }
      else if(!strcmp(pbmid,"P5"))  
      {    strcpy(identifier,"PGMRAW");  return(PGMRAW); }
      else if(!strcmp(pbmid,"P6"))  
      {    strcpy(identifier,"PPMRAW");  return(PPMRAW); }
      else if(check_if_png(identifier))  
      {    strcpy(identifier,"PNG");  return(PNG); }
      else if((word1==0xAFAF)&&(word2==0x7453))
      {    strcpy(identifier,"BIORAD");  return(BIORAD); }     
      else if(pcx_version==253)
      {    strcpy(identifier,"IMA");  return(IMA); }   //// put last, often mistaken for p3   
      else                    
      {                             // Check if an IMG by checking its size.
          xsize=word1;
          ysize=word2;
          fsize = filesize(newfilename);

          // Expected file size. Some valid files have a header or tail at end 
          // Assume > 10 pixels in image

          k = 2*sizeof(short int) + xsize*ysize;   
          if((fsize!=0 && k>10) || strcmp(extension,"img")==SAME)
          {
            if(((double)k/(double)fsize >=.95) && ((double)k/(double)fsize <=1.1))
            {    strcpy(identifier,"IMG"); return(IMG); }    // 8-bit bitmap
            else 
            if(((double)k/(double)(fsize*8)>=.95)&&((double)k/(double)(fsize*8)<=1.1))
            {    strcpy(identifier,"IMM"); return(IMM); }    // Monochrome bitmap
          }
      }

      // Check if there is an extra 128-byte header added by Mac-TCP.                            
      for(k=0;k<128;k++) header[k]=header[k+128];
      g.tif_skipbytes+=128;
   }

   ////  Eliminating Mac header didn't help so reset tif_skipbytes to 0.
   g.tif_skipbytes=0;   
   sprintf(tempstring,"%s\nUnknown image file type",filename);
   message(tempstring, ERROR);
   strcpy(identifier,"NONE");  
   return(-1); 
}


//--------------------------------------------------------------------------//
// strupr                                                                   //
//--------------------------------------------------------------------------//
char *strupr(char *ss)
{
   int kk;
   for(kk=0; kk<(int)strlen(ss); kk++) ss[kk]=toupper(ss[kk]);
   return ss;
}


//--------------------------------------------------------------------------//
// strlwr                                                                   //
//--------------------------------------------------------------------------//
char *strlwr(char *ss)
{
   int kk;
   for(kk=0; kk<(int)strlen(ss); kk++) ss[kk]=tolower(ss[kk]);
   return ss;
}


//--------------------------------------------------------------------------//
//  remove_terminal_cr                                                      //
//--------------------------------------------------------------------------//
void remove_terminal_cr(char *ss)
{
    int pos = max(0, (int)(strlen(ss))-1);
    if(ss[pos]=='\n') ss[pos]=0;
}


//--------------------------------------------------------------------------//
//  filesize                                                                //
//--------------------------------------------------------------------------//
int filesize(char *filename)
{
     struct stat dstatus;
     stat(filename, &dstatus);
     return dstatus.st_size;
}  
 
//--------------------------------------------------------------------------//
//  is_dir                                                                  //
//--------------------------------------------------------------------------//
int is_dir(char *filename)
{
     struct stat dstatus;
     int mode;
     stat(filename, &dstatus);
     mode = dstatus.st_mode;
     if(S_ISDIR(mode)) return 1;
     return 0;
}  


//--------------------------------------------------------------------------//
//  recombine_frames - returns 0 unless user clicked Cancel or error        //
//--------------------------------------------------------------------------//
int recombine_frames(int &ino)
{
  char tempstring[FILENAMELENGTH];
  int source[MAXIMAGES];
  int k, count=0, obpp, answer=OK; 
  if(!between(ino, 0, g.image_count-1) || !z[ino].split_frames) return OK;
  count = 0;
  for(k=z[ino].split_frame_start; k<=z[ino].split_frame_end; k++) source[count++]=k;
  if(count <= 1)
  {   sprintf(tempstring,"Only 1 frame left\n(Originally %d frames)", z[ino].oframe_count);
      return message(tempstring, WARNING);
  }
  sprintf(tempstring,"Image was originally %d frames.\nThere are %d frames remaining.\nShould they be\nrecombined before saving?", z[ino].oframe_count,count);
  if((answer = message(tempstring, YESNOQUESTION)) == YES) 
  {   if(check_image_list(source, count, 1)) return CANCEL; // c.i.l displays err mesage
      obpp = z[source[0]].bpp;
  }else return answer;
  for(k=source[0]; k<=source[count-1]; k++)
  {   if(!between(k, 0, g.image_count-1) || !z[k].split_frames)
      {    message("Image not found\nCan't combine original frames", ERROR);
           return CANCEL;
      }
      switchto(k);
      change_image_depth(k, 24, 1);  // Convert to 24 so common colormap can be generated
      rebuild_display(k);
      redraw(k);
  }
  
  //// This creates new image & changes ci
  if(combine_image_frames(z[ino].xpos+10, z[ino].ypos+10, z[ino].shell, 
       z[ino].window_border, source, count) == OK) 
  {    ino = combinedino = ci;
       switchto(ino);
       change_image_depth(ino, obpp, 1);
       z[ino].format = GIF89;
       rebuild_display(ino);
       redraw(ino);
       setimagetitle(ino, "Combined");
  }else
  {    message("Unable to combine frames", ERROR); 
       return CANCEL;
  }
  
  for(k=source[0]; k<=source[count-1]; k++)
  {    if(between(k, 0, g.image_count-1) && z[k].split_frames)
       {   switchto(k);                    
           change_image_depth(k, obpp, 1);
           rebuild_display(k);
           redraw(k);
       }
  }
  switchto(ino);
  return OK; 
}

//--------------------------------------------------------------------------//
//  array position  - where does char cc first occur in string              //
//--------------------------------------------------------------------------//
int array_position(char *ss, char cc)
{ 
    int kk;
    for(kk=0; kk<(int)strlen(ss); kk++) if(ss[kk]==cc) break;
    return kk;
}

//--------------------------------------------------------------------------//
//  array position_set - replace all char cc with value, assume 8 bits      //
//--------------------------------------------------------------------------//
void set_array_position(char *ss, char cc, int value)
{ 
    int kk;
    for(kk=0; kk<(int)strlen(ss); kk++) if(ss[kk]==cc) ss[kk]=value;
}

//--------------------------------------------------------------------------//
//  remove extension - removes a specific extension                         //
//  Assume extension is the last one in the string                          //
//--------------------------------------------------------------------------//
void remove_extension(char *filename, char *extension)
{
   unsigned long int start = (unsigned long int) filename;
   char *pos = strstr(filename, extension);
   unsigned long int end = (unsigned long int) pos;
   if(end > start) filename[end-start] = 0;
}

//--------------------------------------------------------------------------//
//  remove all extensions - removes everything after first '.'              //
//--------------------------------------------------------------------------//
void remove_all_extensions(char *filename)
{
   int kk;
   for(kk=0; kk<(int)strlen(filename); kk++) 
        if(filename[kk]=='.'){ filename[kk]=0; break;}
   return;
}

//--------------------------------------------------------------------------//
//  find extension start - returns position of first extension in filename  //
//--------------------------------------------------------------------------//
int find_extension_start(char *filename)
{
   int kk, pos=0;
   for(kk=0; kk<(int)strlen(filename); kk++) if(filename[kk]=='.'){pos=kk; break;}
   return pos;
}

//--------------------------------------------------------------------------//
//  find last two extensions returns last and 2nd last file extension       //
//  Puts the extensions in last and second & returns position of last one   //
//  Returns empty string if not found.                                      // 
//  Caller must allocate last[] and second[]                                //
//--------------------------------------------------------------------------//
int find_last_two_extensions(char *filename, char *last, char *second)
{
   int pos=0, oldpos=0, kk, oldkk=0;
   int len = (int)strlen(filename);
   for(kk=0; kk<(int)strlen(filename); kk++) 
   {    if(filename[kk]=='.'){ pos=kk; oldpos=oldkk; oldkk=kk;}
   }
   if(pos){ strncpy(last, filename+pos, len-pos); last[len-pos]=0; }
   else last[0] = 0;
   if(oldpos){ strncpy(second, filename+oldpos, pos-oldpos); second[pos-oldpos]=0; }
   else second[0] = 0;
   return pos;
}

//--------------------------------------------------------------------------//
//  get extension - finds all extensions and puts it in result              //
//--------------------------------------------------------------------------//
void get_extension(char *filename, char *result)
{
   int kk, pos=0;
   if (result==NULL) return;
   result[0] = 0;

   //// start at 1 in case it is a dot file
   //// this willl happen if called from specialcharacter

   for(kk=1; kk<(int)strlen(filename); kk++) 
   {   if(filename[kk]=='.' && filename[kk-1] != '/')
       { pos=kk; break;}
   }
   if(result != NULL) strcpy(result, filename+pos);
   return;
}




