//--------------------------------------------------------------------------//
// xmtnimage6.cc                                                            //
// Help system                                                              //
// Latest revision: 11-07-2013                                              //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//
 
#include "xmtnimage.h"

extern Globals     g;
extern Image      *z;
extern int         ci;
extern PrinterStruct printer;
static int deconvolution_alg_selection=1;
static int deconvolution_psf_selection=1;
static int decon_img_ino    = 1;
static int decon_psf_ino    = 2;
static int decon_iterations = 3;
static int decon_psize = 21;
static double decon_factor  = 65536.0;
 
//--------------------------------------------------------------------------//
// sort                                                                     //
// Returns a sorted index of input array of strings with 'count'            //
// elements. Does not sort the input array itself. (Comb sort)              //
//--------------------------------------------------------------------------//
void sort(char *item[], int sorted[], int count)
{
    int k,gap,iswitch,ii,jj;
    gap = count;
    for(k=0;k<count;k++) sorted[k] = k;
    do
    {  gap = max((int)(gap/1.3), 1);
       iswitch = 0;
       for(ii=0;ii<count-gap;ii++)
       {  jj = ii + gap;
          if(strcmp(item[sorted[ii]], item[sorted[jj]]) > 0)
          {  swap(sorted[ii],sorted[jj]);
             iswitch++;
          }   
       }
    } while((iswitch!=0)||(gap!=1));   
}      


//--------------------------------------------------------------------------//
// sort                                                                     //
// Returns a sorted index of input array of ints with 'count'               //
// elements. Does not sort the input array itself. (Comb sort)              //
//--------------------------------------------------------------------------//
void sort(int item[], int sorted[], int count)
{
    int k,gap,iswitch,ii,jj;
    gap = count;
    for(k=0;k<count;k++) sorted[k] = k;
    do
    {  gap = max((int)(gap/1.3), 1);
       iswitch = 0;
       for(ii=0;ii<count-gap;ii++)
       {  jj = ii + gap;
          if(item[sorted[ii]] > item[sorted[jj]])
          {  swap(sorted[ii],sorted[jj]);
             iswitch++;
          }   
       }
    } while((iswitch!=0)||(gap!=1));   
}      



//--------------------------------------------------------------------------//
// sort (for doubles) - should replace with template                        //
//--------------------------------------------------------------------------//
void sort(double item[], int sorted[], int count)
{
    int k,gap,iswitch,ii,jj;
    gap = count;
    for(k=0;k<count;k++) sorted[k] = k;
    do
    {  gap = max((int)(gap/1.3), 1);
       iswitch = 0;
       for(ii=0;ii<count-gap;ii++)
       {  jj = ii + gap;
          if(item[sorted[ii]] > item[sorted[jj]])
          {  swap(sorted[ii], sorted[jj]);
             iswitch++;
          }   
       }
    } while((iswitch!=0)||(gap!=1));   
}      


//--------------------------------------------------------------------------//
// help                                                                     //
// Reads 'imal.hlp' and provides context-sensitive help.                    //
// Puts text in a list box.                                                 //
//--------------------------------------------------------------------------//
void help(int topic)
{
  if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
  static char **help_info;                                    
  static listinfo *help_list;
  static char *help_title;
  help_title = new char[100];
  help_info = new char*[4096];     // Maximum of 4096 lines of help
  help_list = new listinfo;
  int tno;
  int intopic=0;
  FILE *fp;
  char tempstring[2*FILENAMELENGTH];

  //// Variables for list box
  int listcount;
  int firstitem=0;
  int helptopic=34;

  if(topic==0) topic=1;
  if(topic<=-1)
  {   if(g.imageatcursor>0) { show_notes(ci); return; }
      else topic = 1;
  }
  sprintf(tempstring, "%s/%s",g.helpdir,"imal.hlp");
  if ((fp=fopen(tempstring,"r")) == NULL)
  {   sprintf(tempstring, "%s/%s",g.homedir,"imal.hlp");
      if ((fp=fopen(tempstring,"r")) == NULL)
      {   if((fp=fopen("imal.hlp","r")) == NULL)
          {   sprintf(tempstring, "Can't find help file\n(Should be: %s/imal.hlp)",g.helpdir);
              message (tempstring,3);
              return;
          }
      }
  }
  tno=0;
  intopic=0;
  while(!feof(fp))             // Count no. of lines in topic
  {  fgets(tempstring,85,fp); 
     if(isdigit(tempstring[0])) tno = atoi(tempstring);
     if(tno==topic) intopic=1; 
     if((intopic) && (tno) && (tno!=topic)) break;
  }
  intopic=0;
  tno=0;

  fseek(fp,0,SEEK_SET);
  listcount = 0;
  while(!feof(fp))
  {  fgets(tempstring,85,fp); 
     if(isdigit(tempstring[0])) tno = atoi(tempstring);
     if(tno==topic) intopic=1;
     if((intopic) && (tno) && (tno!=topic)) break;
     if(intopic)
     {   tempstring[80] =0;
         remove_terminal_cr(tempstring);
         help_info[listcount] = new char[100];
         help_info[listcount][0]=0;
         if(strlen(tempstring)) strcpy(help_info[listcount], tempstring);
         else strcpy(help_info[listcount]," ");
         listcount++;
     }
     if(listcount>1023) break;
  }
  fclose(fp);

  if(listcount)
  {  strcpy(help_title, help_info[0]);
     help_list->title = help_title;
     help_list->info  = help_info;
     help_list->count = listcount;
     help_list->itemstoshow = min(10, listcount);
     help_list->firstitem   = firstitem;
     help_list->wantsort    = 0;
     help_list->wantsave    = 0;
     help_list->helptopic   = helptopic;
     help_list->allowedit   = 0;
     help_list->selection   = &g.crud;
     help_list->width       = 0;
     help_list->transient   = 1;
     help_list->wantfunction = 0;
     help_list->autoupdate   = 0;
     help_list->deletebutton = 0;
     help_list->clearbutton  = 0;
     help_list->exportbutton = 0;
     help_list->highest_erased = 0;
     help_list->x = 0;
     help_list->y = 0;
     help_list->f1 = null; // use f1 to perform an action when item is selected
     help_list->f2 = null; // use f2 to update list 
     help_list->f3 = null;
     help_list->f4 = delete_list;
     help_list->f5 = null;
     help_list->f6 = null;
     help_list->listnumber = 0;
     list(help_list);
  }
}


//--------------------------------------------------------------------------//
// show_notes                                                               //
//--------------------------------------------------------------------------//
void show_notes(int ino)
{
  FILE *fp;
  static char filename[2*FILENAMELENGTH];
  sprintf(filename, "%s/notes/%s", g.homedir, basefilename(z[ino].name));
  int fsize = filesize(filename);
  if(z[ino].notes == NULL)
  {   
      z[ino].notes = new char[65536];
      if(z[ino].notes == NULL)
      {   message(g.nomemory); 
          return; 
      }
      z[ino].notes[0] = 0;
      if(fsize==0 || access(filename, F_OK)) // access = 0 if file exists
      {
          message("No image notes found"); 
      }else
      {   if((fp=fopen(filename,"r")) == NULL)
          {   error_message(filename, errno);
              return;
          }
          if(fread(z[ino].notes, fsize, 1, fp) == 0)
          {   message("Notes string had zero length"); 
              delete z[ino].notes; 
              z[ino].notes = NULL; 
              fclose(fp);
              return; 
          }
          fclose(fp);
      }
  }
  edit("Image notes", "Image notes", z[ino].notes, 20, 80, 0,0, 65536, 
      70, 0, filename, null, NULL);
}


//--------------------------------------------------------------------------//
// deconvolution                                                            //
// Richardson-Lucy deconvolution                                            //
//--------------------------------------------------------------------------//
void deconvolution(void)
{
   static Dialog *dialog;
   int j,k;
   if(memorylessthan(16384)){  message(g.nomemory,ERROR); return; } 
 
   dialog = new Dialog;
   if(dialog==NULL){ message(g.nomemory); return; }
   dialog->helptopic = 77;  
   strcpy(dialog->title,"Deconvolution");
         
   ////--Radio buttons--////
 
   strcpy(dialog->radio[0][0],"Algorithm");
   strcpy(dialog->radio[0][1],"Richardson-Lucy");             
   strcpy(dialog->radio[0][2],"Richardson-Lucy");             

   ////-----Boxes-----////

   strcpy(dialog->boxes[0],"Parameters");
   strcpy(dialog->boxes[1],"Image to change");
   strcpy(dialog->boxes[2],"Image no. for PSF");
   strcpy(dialog->boxes[3],"No. of iterations");

   dialog->boxtype[0]=LABEL;
   dialog->boxtype[1]=INTCLICKBOX; dialog->boxmin[1]=1; dialog->boxmax[1]=g.image_count-1;
   dialog->boxtype[2]=INTCLICKBOX; dialog->boxmin[2]=1; dialog->boxmax[2]=g.image_count-1;
   dialog->boxtype[3]=INTCLICKBOX; dialog->boxmin[3]=1; dialog->boxmax[3]=100;

   sprintf(dialog->answer[1][0], "%d", decon_img_ino);
   sprintf(dialog->answer[2][0], "%d", decon_psf_ino);
   sprintf(dialog->answer[3][0], "%d", decon_iterations);
   
   dialog->radioset[0] = deconvolution_alg_selection;
   dialog->radiono[0] = 2;
   for(j=0;j<10;j++) for(k=0;k<20;k++) dialog->radiotype[j][k]=RADIO;
   dialog->noofradios = 1;
   dialog->noofboxes = 4;
   dialog->want_changecicb = 0;
   dialog->f1 = deconvolution_check;
   dialog->f2 = null;
   dialog->f3 = deconvolution_cancel;
   dialog->f4 = deconvolution_finish;
   dialog->f5 = null;
   dialog->f6 = null;
   dialog->f7 = null;
   dialog->f8 = null;
   dialog->f9 = null;
   dialog->transient = 1;
   dialog->wantraise = 0;
   dialog->radiousexy = 0;
   dialog->boxusexy = 0;
   dialog->radioextrawidth = 0; //// this must be set or it screws up
   dialog->width = 0;  // calculate automatically
   dialog->height = 0; // calculate automatically
   dialog->ptr[4] = 0;
   strcpy(dialog->path,".");
   g.getout=0;
   dialog->message[0]=0;      
   dialog->busy = 0;
   dialogbox(dialog);

}

//--------------------------------------------------------------------------//
// deconvolution_check - richardson-lucy deconvolution                      
// WH Richardson, J Opt Soc Am 62(1), 55-59 (1970)
// "Bayesian-based iterative method of image restoration"
//--------------------------------------------------------------------------//
void deconvolution_check(dialoginfo *a, int radio, int box, int boxbutton)
{
  a=a; radio=radio; box=box; boxbutton=boxbutton;
  int ino, tagcount;
  Widget www, scrollbar;
  int bpp, i,j,xsize,ysize,psize, k, kk, n, oo, oldci, ncolors;
  int rr,gg,bb;
  int iteration;
  double v;
  g.getout = 0;
  char title[256];
//  char tempstring[1024];

  //// shut up the compiler
  deconvolution_alg_selection = a->radioset[0];
  deconvolution_psf_selection = a->radioset[1];

  if(radio != -2) return; // Radio is -2 if clicked Accept

  decon_img_ino    = atoi(a->answer[1][0]);
  decon_psf_ino    = atoi(a->answer[2][0]);
  decon_iterations = atoi(a->answer[3][0]);
  ino = oldci = decon_img_ino;


  // pij = point spread function = frac of light from j observed at i
  // uj =  deconvoluted pixel value at j from previous iteration
  // hi =  observed value at i
  //                      
  //                  ___  h   p
  //  w (t+1) = w (t) \     i   ij
  //   j         j    /   --------
  //                 /___     d
  // where                      i
  //            ___
  //  d (t+1) = \    p   u
  //   i        /__   ij  j
  //
  // It is easier to conceptualize this as a series of convolutions.
  //              _            _
  //             |     h        |
  // w     =  w  |   -----  * p |
  //  t+1      t |    w * p     |
  //             |_    t       _|
  //             
  //   where * is convolution operator
  //   h = observed blurred image
  //   p = psf kernel
  //   w = estimate of deconvoluted image
  //   multiplications & divisions are all pixelwise
  // The quotient is a fraction between 0 and 1
  // All pixels are scaled to [0,1]
  // Must be represented as doubles or noise starts to appear after
  //   several iterations
  // Matlab version
  //    for i= 1:max_iterations
  //       estimate = conv2(w, p, 'same');
  //       blur = h / estimate;
  //       errror = conv2(blur, p, 'same'); 
  //       w = w * errror;
  //    end
  // The code below can probably be speeded up


   switchto(decon_img_ino);      //// the image(s) to deconvolute, now = ci
   oldci = ci;                   //// original image
   psize = min(z[decon_psf_ino].xsize,z[decon_psf_ino].ysize);
   decon_psize = psize;
 
   if(z[decon_psf_ino].xsize != z[decon_psf_ino].ysize){ message("PSF is not square"); return; }
   if(z[decon_psf_ino].bpp != z[ci].bpp){ message("PSF has wrong pixel depth"); return; }
   if(z[decon_psf_ino].colortype != z[ci].colortype){ message("PSF has wrong color type"); return; }
   if(z[ci].colortype == INDEXED){ message("Cannot deconvolute indexed-\ncolor images"); return; }
   if(z[decon_psf_ino].tagged){ message("PSF image is tagged, cannot deconvolute it"); return; }
   if(decon_psf_ino == ci){ message("Cannot deconvolute PSF image"); return; }
 
   if(z[ci].colortype == COLOR)
   {    ncolors = 3; 
        decon_factor = (double)g.maxgreen[z[ci].bpp];
   }else
   {    ncolors = 1;
        decon_factor = g.maxvalue[z[ci].bpp];
   }

   progress_bar_open(www, scrollbar,(char*)"Progress");

   tagcount = 0;  
   for(kk=1; kk<g.image_count; kk++) if(z[kk].tagged==1) tagcount++;
   for(kk=1; kk<g.image_count; kk++)
   {   if((z[kk].tagged==1) || (kk==oldci))
       {    
              switchto(kk);
              ino = ci;
              bpp = z[ino].bpp;
              z[ino].hit = 1;
              xsize = z[ci].xsize;
              ysize = z[ci].ysize;
              bpp = z[ci].bpp;
              oo  = g.off[bpp];
              array<double> w(xsize, ysize);    /// old w
              array<double> w2(xsize, ysize);   /// new w
              array<double> blur(xsize, ysize); /// new w
              array<double> p(psize, psize);    /// psf
              array<double> h(xsize, ysize);    /// original
              for(n=0; n<ncolors; n++)
              {
                      for(j=0; j<ysize; j++)
                      for(i=0; i<xsize; i++)
                      {     
                            if(ncolors==3) 
                            {      RGBat(z[ino].image[0][j]+i*oo, bpp, rr, gg, bb);
                                   switch(n)
                                   {     case 0: v = (double)rr; break;
                                         case 1: v = (double)gg; break;
                                         case 2: v = (double)bb; break;
                                   }
                            }else
                            {      v = (double)(pixelat(z[ino].image[0][j]+i*oo,bpp));
                            }
                            v = v / decon_factor;
                            w.p[j][i] = v;
                            w2.p[j][i] = 0;
                            blur.p[j][i] = 0;
                            h.p[j][i] = v;
                      }     
                      for(j=0; j<psize; j++)
                      for(i=0; i<psize; i++)
                      {   
                            v = (double)(pixelat(z[decon_psf_ino].image[0][j]+i*oo,bpp));
                            v = v / decon_factor;
                            p.p[j][i] = v;
                      }     
        
                      for(iteration = 0; iteration < decon_iterations; iteration++)
                      {     
                            sprintf(title, "Image %d Iteration ",ino);
                            if(ncolors==3) 
                            {      switch(n)
                                   {     case 0: strcat(title, "Red"); break;
                                         case 1: strcat(title, "Green"); break;
                                         case 2: strcat(title, "Blue"); break;
                                   }
                            }

                            progress_bar_update_custom(scrollbar, iteration, decon_iterations, (char*) title);

                            //// garbage to stop motif from throwing up a blank message box
                            usleep(100);
                            for(k=0;k<1000;k++)  check_event();
                            usleep(100);

                            //// calculate the denominator  (w conv p)        
                            //// this step needs 2 buffers
                            convolute(w.p, p.p, w2.p);
                            if(g.getout) break;

                            //// divide original observed by denominator, put relative blur in ino
                            divide(h.p, w2.p, blur.p);
                            if(g.getout) break;
        
                            //// convolute the quotient by p
                            //// this step needs 2 buffers, re-use w2 to store error
                            convolute(blur.p, p.p, w2.p);
                            if(g.getout) break;

                            //// multiply the w, element by element, by error, put in old buffer
                            multiply(w.p, w2.p, w.p);
                            if(g.getout) break;

                            //// put w back into ino
                            copy_image_buffer(w.p, ino, n);
                            if(g.getout) break;
                     }

              }
              if(g.getout) break;
              switchto(ino);
              z[ino].hitgray = 0;  // Force remapping of gray levels
              repair(ino);
              redraw(ino);    
              z[ino].touched = 1;
       }
   }
   if(g.getout) g.getout=0;
   progress_bar_close(www);
}


//--------------------------------------------------------------------------//
// convolute                                                                //
// all params are image numbers                                             //
// kernel must be square and same bpp as input and output                   //
// input and output image must be different                                 //
//--------------------------------------------------------------------------//
void convolute(double **w1, double **w2, double **output)
{
  int i,j,k,l,kk,ll,xsize,ysize,psize;
  double aa,bb,value;

  xsize = z[ci].xsize;
  ysize = z[ci].ysize;
  psize = decon_psize;
  for(l=0; l<ysize; l++)  // destination pixel
  for(k=0; k<xsize; k++)
  {  
          value = 0.0;                         
          for(j=0; j<psize; j++)  // psf kernel
          for(i=0; i<psize; i++)
          { 
                aa = w2[j][i];

                ll = l + j - psize/2;
                kk = k + i - psize/2;
                if((ll<0) || (kk<0)) continue;
                if((ll>=ysize) || (kk>=xsize)) continue;
                bb = w1[ll][kk];

                value += aa * bb;
          }                     
          output[l][k] = value;
  }
}


//--------------------------------------------------------------------------//
// divide - divide 2 images pixel by pixel                                  //
// must be same size & bpp                                                  //
//--------------------------------------------------------------------------//
void divide(double **w1, double **w2, double **result)
{
  int j,k,xsize,ysize;
  double aa,bb,value;

  xsize = z[ci].xsize;
  ysize = z[ci].ysize;
  for(j=0; j<ysize; j++)  // destination pixel
  for(k=0; k<xsize; k++)
  {  
         aa = w1[j][k];
         bb = w2[j][k];
         if(bb!=0) value = aa / bb;
         else value = 0;
         result[j][k] = value;
  }
}


//--------------------------------------------------------------------------//
// multiply - multiply 2 images pixel by pixel                              //
// must be same size & bpp                                                  //
//--------------------------------------------------------------------------//
void multiply(double **w1, double **w2, double **result)
{
  int j,k,xsize,ysize;
  double aa,bb,value;

  xsize = z[ci].xsize;
  ysize = z[ci].ysize;

  for(j=0; j<ysize; j++)  // destination pixel
  for(k=0; k<xsize; k++)
  {  
         aa = w1[j][k];
         bb = w2[j][k];
         value = aa * bb;
         result[j][k] = value;
  }
}


//--------------------------------------------------------------------------//
// scaleimage                                                               //
//--------------------------------------------------------------------------//
void scaleimage(int ino, double factor)
{
  int j,k,bpp,oo,xsize,ysize;
  double aa, value;

  bpp = z[ino].bpp;
  oo  = g.off[bpp];
  xsize = z[ino].xsize;
  ysize = z[ino].ysize;
  for(j=0; j<ysize; j++)  // destination pixel
  for(k=0; k<xsize; k++)
  {  
         aa = (double)(pixelat(z[ino].image[0][j]+k*oo,bpp));
         value = aa * factor;
         putpixelbytes(z[ino].image[0][j]+k*oo, (int)(value), 1, bpp, 1);
  }
}


//--------------------------------------------------------------------------//
// copy_image_buffer                                                        //
// buffers must be the same size                                            //
//--------------------------------------------------------------------------//
void copy_image_buffer(double **w, int ino, int n)
{
  int ct,j,k,bpp,oo,xsize,ysize,rr,gg,bb,value;
  ct  = z[ino].colortype;
  bpp = z[ino].bpp;
  oo  = g.off[bpp];
  xsize = z[ino].xsize;
  ysize = z[ino].ysize;
  for(j=0; j<ysize; j++)  // destination pixel
  for(k=0; k<xsize; k++)
  {  
         value = min((int)decon_factor, (int)((double)decon_factor * w[j][k]));
         if(ct==COLOR)
         {    RGBat(z[ino].image[0][j]+k*oo, bpp, rr, gg, bb);
              switch(n)
              {   case 0: rr = min(255, (int) value); break;
                  case 1: gg = min(255, (int) value); break;
                  case 2: bb = min(255, (int) value); break;
              }
              setRGBbytes(z[ino].image[0][j]+k*oo, rr, gg, bb, bpp);
         }else
              putpixelbytes(z[ino].image[0][j]+k*oo, value, 1, bpp, 1);
  }
}

//--------------------------------------------------------------------------//
// deconvolution_cancel                                                     //
//--------------------------------------------------------------------------//
void deconvolution_cancel(dialoginfo *a)
{
   a=a;
   g.getout=1;
}

//--------------------------------------------------------------------------//
// deconvolution_finish                                                     //
//--------------------------------------------------------------------------//
void deconvolution_finish(dialoginfo *a)
{
   a=a;
}
