//--------------------------------------------------------------------------//
// xmtnimage65.cc  png, p2, p3 format                                       //
// Latest revision: 08-28-2023                                              //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"

extern Globals     g;
extern Image      *z;
extern int         ci;

long int simplerand(void);
long uint passwordtoseed(char *password, int version);
long uint seed;
long long unsigned int llseed;
long long int simplerand2(void);
#define ROTL8(x,shift) ((uint8_t) ((x) << (shift)) | ((x) >> (8 - (shift))))

extern "C"
{
#include<png.h>
#include<pngconf.h>
}
void user_error_ptr(void);
void read_png(FILE *fp, unsigned int sig_read);
void png_read_row_callback(png_struct *png_ptr, png_uint_32 row, int pass);
void write_row_callback(png_struct *png_ptr, png_uint_32 row, int pass);
void makesbox(uint8_t *sbox, uint8_t *rev_sbox);


//-------------------------------------------------------------------------//
// check_if_png                                                             //
//-------------------------------------------------------------------------//
int check_if_png(char *identifier)
{
   char test[5] = ".PNG";
   test[0] = 0x89;
   return(!memcmp(identifier, test, 4));
}


//-------------------------------------------------------------------------//
// readpngfile                                                             //
// Fixed to handle libpng 1.6, which was changed for some reason.          //
// Note to self: no more using standard libraries.                         //
// This only supports the most rudimentary form of the PNG format.         //
//-------------------------------------------------------------------------//
int readpngfile(char *filename)
{
  FILE *fp;
  int bpp, ino, ct=0, k, color_type, interlace_type, compression_type, filter_type;
  int num_palette, number_of_passes = 1;
  png_colorp palette;
  //double gamma;
  png_uint_32 w, h;
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
  if (!png_ptr) return ERROR;
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
  {     png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        return ERROR;
  }
  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info)
  {     png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        return ERROR;
  }
  if ((fp=fopen(filename,"r")) == NULL){ message("Nope"); return ERROR; }
//  if (setjmp(png_ptr->jmpbuf))
//  {     png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
//        fclose(fp);
//        return ERROR;
//  }
  png_init_io(png_ptr, fp);
  png_set_read_status_fn(png_ptr, png_read_row_callback);
  png_read_info(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &w, &h, &bpp, &color_type, 
        &interlace_type, &compression_type, &filter_type);
  if( (number_of_passes = png_set_interlace_handling(png_ptr)) <0){ message("Error in png interlace handling"); return ERROR; }
  switch(color_type)
  {     case PNG_COLOR_TYPE_GRAY:
        case PNG_COLOR_TYPE_GRAY_ALPHA:
             ct=GRAY; 
             break;
        case PNG_COLOR_TYPE_PALETTE:
             ct=INDEXED;
             break;
        case PNG_COLOR_TYPE_RGB:
        case PNG_COLOR_TYPE_RGB_ALPHA:
             ct=COLOR;
             bpp=max(bpp, 24);  //// png struct returns wrong bpp
             break;
        default: message("Unsupported PNG type");
        {     png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
              fclose(fp);
              return ERROR;
        }             
  }
  if(!between(bpp,0,48))
  {     png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
        fclose(fp);
        return ERROR;
  }
  if (bpp < 8) png_set_packing(png_ptr);
  //gamma = 1.0; 
  //number_of_passes = png_set_interlace_handling(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY && bpp < 8) png_set_expand(png_ptr);
  if(ct==COLOR) png_set_strip_16(png_ptr);
  png_set_strip_alpha(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
  bpp = 8 * ((bpp + 7) / 8);
  if(newimage(basefilename(filename),0,0,w,h,bpp,ct,1,0,1,PERM,1,0,0)!=OK)
  {     png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
        fclose(fp);
        return ERROR;
  }
  ino = ci;

  //// Make an array of ptrs as libpng expects
  uchar **junk;
  junk = new uchar*[h];
  for(k=0; k<(int)h; k++) junk[k] = z[ino].image[0][k];
  png_read_image(png_ptr, &junk[0]);
  delete[] junk;

  if (ct==INDEXED && PNG_COLOR_MASK_COLOR)
  {     png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette);
        for(k=0;k<256;k++)
        {   g.palette[k].red   = palette[k].red/4;
            g.palette[k].green = palette[k].green/4;
            g.palette[k].blue  = palette[k].blue/4;
        }
  }

  //png_set_gamma(png_ptr, 1.0, 1.0); // Image readers should not set gamma
  png_read_end(png_ptr, end_info);
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

  ////  Swap bytes for color only, not grayscale
  if (ct==COLOR && bpp >= 16 && g.ximage_byte_order == LSBFirst) 
       g.invertbytes = 1;
  fclose(fp);
  return OK;
}


void png_read_row_callback(png_struct *png_ptr, png_uint_32 row, int pass)
{
   png_ptr=png_ptr; row=row; pass=pass;
   // printf("row callback\n");
}



//-------------------------------------------------------------------------//
// writepngfile - can only write entire image                              //
//-------------------------------------------------------------------------//
int writepngfile(char *filename)
{
  if(memorylessthan(16384)){ message(g.nomemory,ERROR); return(NOMEM); }
  FILE *fp;
  png_colorp palette;
  int bpp,png_bpp,k,w,h,ct,color_type;
  int xstart,ystart,xend,yend;
  ct = z[ci].colortype;
  bpp = z[ci].bpp;
  //png_color_8p sig_bit = NULL;
  xstart = z[ci].xpos;
  ystart = z[ci].ypos;
  xend = xstart + z[ci].xsize;
  yend = ystart + z[ci].ysize;
  w = xend - xstart;
  h = yend - ystart;
  if((fp=open_file(filename, "w"))==NULL) return CANTCREATEFILE; 
  png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 
      (void *)NULL, NULL, NULL);
  if (!png_ptr) return ERROR;
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr){ png_destroy_write_struct(&png_ptr, (png_infopp)NULL); return ERROR;  }

// this no longer works
//  if (setjmp(png_ptr->jmpbuf))
//  {   png_destroy_write_struct(&png_ptr, &info_ptr);
//      fclose(fp);
//      return ERROR;
//  }
  png_bpp = 8;

  png_init_io(png_ptr, fp);
  png_set_write_status_fn(png_ptr, write_row_callback);
  switch(ct)
  {     case GRAY: 
             png_bpp = bpp;
             color_type = PNG_COLOR_TYPE_GRAY; 
             break;
        case INDEXED: 
             png_bpp = 8;
             color_type = PNG_COLOR_TYPE_PALETTE; 
             break;
        case COLOR: 
             png_bpp = 8;
             color_type = PNG_COLOR_TYPE_RGB; 
             break;
        default: color_type = PNG_COLOR_TYPE_RGB; break;
  }
  png_set_IHDR(png_ptr, info_ptr, w, h, png_bpp, color_type, PNG_INTERLACE_NONE,
       PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  palette = (png_colorp)png_malloc(png_ptr, 256 * sizeof (png_color));
  if(ct!=GRAY)
  {    for(k=0;k<256;k++)
       {    palette[k].red   = z[ci].palette[k].red * 4;
            palette[k].green = z[ci].palette[k].green * 4;
            palette[k].blue  = z[ci].palette[k].blue * 4;
       }
       png_set_PLTE(png_ptr, info_ptr, palette, 256);
  }

  if(ct == COLOR) 
  { 
  // These all cause crash for unknown reason
  //       sig_bit->red = g.maxred[bpp];
  //       sig_bit->green = g.maxgreen[bpp];
  //       sig_bit->blue = g.maxblue[bpp];
  //       png_set_sBIT(png_ptr, info_ptr, sig_bit);
  }
  //if(ct == GRAY)  
  //{     sig_bit->gray = bpp;
  //      png_set_sBIT(png_ptr, info_ptr, sig_bit);
  //}

  if (color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_RGB_ALPHA)
       png_set_bgr(png_ptr);
  png_write_info(png_ptr, info_ptr);
  if(bpp==32) png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);
  if (bpp >= 16 && g.ximage_byte_order == LSBFirst) png_set_swap(png_ptr);

  //// Make an array of ptrs as libpng expects
  uchar **junk;
  junk = new uchar*[h];
  for(k=0; k<(int)h; k++) junk[k] = z[ci].image[0][k];

  png_write_image(png_ptr, junk);
  delete[] junk;
  png_write_end(png_ptr, info_ptr);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  close_file(fp);  
  return OK; 
}

void write_row_callback(png_struct *png_ptr, png_uint_32 row, int pass)
{
  png_ptr=png_ptr; row=row; pass=pass;
     /* put your code here */
}


//-------------------------------------------------------------------------//
// readp2file                                                              //
// simple, not particularly strong password encrypted format               //
//-------------------------------------------------------------------------//
int readp2file(char *filename)
{   
   if(memorylessthan(16384)){ message(g.nomemory,ERROR); return NOMEM; }
   FILE *fp;
   int i, i2, j, colortype;
   int v;
   int bpp, frames, xsize, ysize;
   char tempstring[1024];
   char identifier[1024];
   char name[1024];
   char hashed_password[1024];

   //// placeholder for future hashed password in file
   strcpy(hashed_password, "#dummy password here#");
   strcpy(hashed_password, "#");     
   for(i=0; i<1024; i++) tempstring[i] = 0;                 

   uchar *add;
   char defaultfilename[64];
   strcpy(defaultfilename, "Image"); // keep compiler quiet
   frames = 1;

   if ((fp=fopen(filename,"r")) == NULL)
   {   error_message(filename, errno);
       return(NOTFOUND);
   }

   //// crude read protection
   //// don't use linux built-in hash algorithm, it could be depreciated
   //// at any time making all files unreadable

   //// use same password for all p2 files
   //// unless user turns this feature off

   if(!strlen(g.p2password) || g.password_mode==0) /// no password or always ask
   {   
          if(g.password_mode==0)
                sprintf(tempstring, "Password for\n%s:", filename);
          else
                strcpy(tempstring, "Password for p2 files:");
          getstring(tempstring, g.p2password, 1024, 78);
   }

   seed = passwordtoseed(g.p2password, 2);
   fread(tempstring, 1024, 1, fp);
   fread(z[ci].palette, 768, 1, fp);
   sscanf(tempstring,"%s %s %d %d %d %d %d\n", 
          identifier,
          name,
          &xsize,
          &ysize,
          &bpp,
          &frames,
          &colortype);

   if(newimage(defaultfilename, 0, 0, xsize, ysize, bpp,
       colortype, frames, g.want_shell, 1, PERM, 
       g.window_border,1,0) != OK){ fclose(fp);  return NOMEM; }


   //// decrypt file
   for(j=0;j<z[ci].ysize;j++)
   {  
      for(i=0,i2=0; i<z[ci].xsize; i++,i2+=g.off[bpp])
      {   fread(&v, sizeof(int), 1, fp);  
          v -= simplerand();
          add = z[ci].image[0][j]+i2; 
          putpixelbytes(add, v, 1, bpp, 1);
      }
   }     
   fclose(fp);
   return OK;
}


//-------------------------------------------------------------------------//
// writep2file - can only write entire image                               //
// simple, easily cracked password encryption, obscures image f/ casual use//
//-------------------------------------------------------------------------//
int writep2file(char *filename)
{
  int bpp,i,i2,j;
  int v;
  FILE *fp;
  char tempstring[1024];
  strcpy(z[ci].name, basefilename(filename));
  if((fp=open_file(filename, "w"))==NULL) return CANTCREATEFILE; 

  bpp = z[ci].bpp;
  memset(tempstring, 32, 1024);
  sprintf(tempstring, "%s %s %d %d %d %d %d\n", 
       "p2",
       z[ci].name, 
       z[ci].xsize,
       z[ci].ysize,
       z[ci].bpp,
       z[ci].frames,
       z[ci].colortype);

  //// first 1024 bytes are reserved for image params
  //// second 1024 bytes are reserved for future hashed password

  fwrite(tempstring, 1024, 1, fp);  
  fwrite(z[ci].palette, 768, 1, fp);

  //// use same password for all p2 files
  //// unless user turns this feature off

  if(!strlen(g.p2password) || g.password_mode==0) /// no password or always ask
  {  
         getstring("Set password for encrypting image\nWarning:keep a record of the password\nor image will be unrecoverable", g.p2password, 1024, 78);
  }
  seed = passwordtoseed(g.p2password, 2);

  //// quick and easy way of encrypting file to a password
  for(j=0;j<z[ci].ysize;j++)
  {  
     for(i=0,i2=0; i<z[ci].xsize; i++,i2+=g.off[bpp])
     {   v = pixelat(z[ci].image[0][j]+i2, bpp);
         v += simplerand();
         fwrite(&v, sizeof(int), 1, fp);  
     }
  }     
  close_file(fp);  
  return OK;
}


//-------------------------------------------------------------------------//
// simplerand - random number generator period 2^31                        //
//-------------------------------------------------------------------------//
long int simplerand(void)
{

   // arxiv.org/abs/2001.05304
   // xb = (a  xn-1 + c) mod m
   // linear congruential pseudorandom number generator

   static long int m = 2147483648;  // 2^31 a power of 2 modulus
   static long int a = 1103515245;  // multiplier
   static long int c = 12345;
   seed = (a * seed + c) % m;
   return(seed);
}

//-------------------------------------------------------------------------//
// passwordtoseed - derive numerical key from string password              //
//-------------------------------------------------------------------------//
long uint passwordtoseed(char *password, int version)
{
   int kk;
   long uint answer = 0;
   for(kk=0; kk<(int)strlen(password); kk++)
   {
        //// 7 lower bits at a time
       switch (version)
       {
           case 2: answer += (password[kk]&0x7f) << (7*kk);
                   break;
           case 3: 
                   answer += (password[kk]&0x7f) << (7*kk);
                   break;
           default: message("An error occurred");
       }
   }
   return answer;
}


//-------------------------------------------------------------------------//
// readp3file                                                              //
// improved stream cipher fast but only moderate security                  //
//-------------------------------------------------------------------------//
int readp3file(char *filename)
{   
   if(memorylessthan(16384)){ message(g.nomemory,ERROR); return NOMEM; }
   FILE *fp;
   int i, i2, j, colortype;
   long unsigned int originalcode = 12345;
   long unsigned int code = 0;
   long unsigned int salt = 1;
   long unsigned int oldsalt;
   int v;
   int bpp, frames, xsize, ysize;
   char tempstring[1024];
   char identifier[1024];
   char name[1024];
   char hashed_password[1024];

   //// placeholder for future hashed password in file
   strcpy(hashed_password, "#dummy password here#");
   strcpy(hashed_password, "#");     
   for(i=0; i<1024; i++) tempstring[i] = 0;                 

   uchar *add;
   char defaultfilename[64];
   strcpy(defaultfilename, "Image"); // keep compiler quiet
   frames = 1;

   if ((fp=fopen(filename,"r")) == NULL)
   {   error_message(filename, errno);
       return(NOTFOUND);
   }

   if(!strlen(g.p3password) || g.password_mode==0) /// no password or always ask
   {   
          if(g.password_mode==0)
                sprintf(tempstring, "p3 Password for\n%s:", filename);
          else
                strcpy(tempstring, "Password for p3 files:");
          getstring(tempstring, g.p3password, 1024, 78);
   }

   fread(&salt, 8, 1, fp);              //// 8 byte salt
   fread(&code, 8, 1, fp);              //// 8 byte code
   fread(tempstring, 1024, 1, fp);      //// image params

   seed = passwordtoseed(g.p3password, 3);
   oldsalt = salt;
   salt ^= simplerand() / 27;           //// simplerand changes seed
   seed = oldsalt+seed;
   code ^= simplerand() / 27;          //// simplerand changes seed

   for(i=0; i<1024; i++)
         tempstring[i] ^= simplerand() % 256; //// xor

   sscanf(tempstring,"%s %s %d %d %d %d %d\n", 
          identifier,
          name,
          &xsize,
          &ysize,
          &bpp,
          &frames,
          &colortype);

   //// To test the encryption, comment this out and set the known
   //// image params below. Then use XOR masking with a test image. 
   //// If passwords are different, there should be no identifiable 
   //// parts of the image. Only if the correct password is used to 
   //// generate a test image should any information be visible.

   if(strcmp(identifier, "p3") || (code != originalcode))
   {   sprintf(tempstring, "Incorrect password or corrupt p3 file\n%s",filename);
       message(tempstring, ERROR);
       fclose(fp);
       return(ERROR);
   }      

/*
xsize=142;
ysize=147;
bpp=24;
colortype=COLOR;
*/
/*
xsize=2421;
ysize=3645;
bpp=24;
colortype=COLOR;
*/

   if(newimage(defaultfilename, 0, 0, xsize, ysize, bpp,
       colortype, frames, g.want_shell, 1, PERM, 
       g.window_border,1,0) != OK){ fclose(fp);  return NOMEM; }


   //// decrypt file
   for(j=0;j<z[ci].ysize;j++)
   {  
      for(i=0,i2=0; i<z[ci].xsize; i++,i2+=g.off[bpp])
      {   fread(&v, sizeof(int), 1, fp);  
          v ^= simplerand() / 27;   //// xor
          add = z[ci].image[0][j]+i2; 
          putpixelbytes(add, v, 1, bpp, 1);
      }
   }     
   fclose(fp);
   return OK;

}

//-------------------------------------------------------------------------//
// writep3file - can only write entire image                               //
// There is no point in compressing an encrypted file, but the software    //
// doesn't stop you                                                        //
//-------------------------------------------------------------------------//
int writep3file(char *filename)
{
  long unsigned int code = 12345;
  long unsigned int salt = 1;
  int bpp,i,i2,j;
  int v;    //// must be 32 bit int
  FILE *fp;
  char tempstring[1024];
  strcpy(z[ci].name, basefilename(filename));
  if((fp=open_file(filename, "w"))==NULL) return CANTCREATEFILE; 

  bpp = z[ci].bpp;

  //// fill unused part of tempstring with junk
  srand((int)time((time_t *)0));
  memset(tempstring, rand()%0xff, 1024);
  sprintf(tempstring, "%s %s %d %d %d %d %d\n", 
       "p3",
       z[ci].name, 
       z[ci].xsize,
       z[ci].ysize,
       z[ci].bpp,
       z[ci].frames,
       z[ci].colortype);
  tempstring[1024]=0;

  //// use same password for all p3 files
  //// unless user turns this feature off

  //// Don't use linux built-in hash algorithm, it could be depreciated
  //// at any time making all files unreadable.

  //// Use same password for all p3 files
  //// unless user turns this feature off
  //// This would make it easy to deduce the key, so we add part of the
  //// image to the key to salt it.

  //// Note: this is not intended to be a secure encryption.
  //// It is useful only to conceal the image from a casual observer.
  //// It should not be relied on as a form of security.
  //// A competent cryptographer could probably break this in four hours
  //// on a desktop PC.

  //// encoding
  //// 0. get password from user and convert to seed
  //// 1. encrypt salt using original seed and write to file
  //// 2. add encrypted salt to seed
  //// 3. encrypt image with new seed and save

  //// decoding uses a bootstrapping trick to get the actual key,
  //// which is a combination of the user password and part of the image.
  //// We wish to avoid any linux-specific or hardware-specific functions
  //// because Linux has a long history of changing essential
  //// things for the sake of change, which would make our files
  //// unreadable. So instead of /dev/urandom we will use our own PRNG 
  ////
  //// 0. get password from user and convert to seed
  //// 1. read encrypted salt from file 
  //// 2. decrypt using original seed and add to seed
  //// 3. read and decrypt image with new seed 

  //// Add 6 bytes from the image to the password
  //// Assume image is at least 6 bytes long
  //// A longer password is better
  for(i=0; i<6; i++) salt *= max(1, z[ci].image[0][0][i] & 0xff); 

  if(!strlen(g.p3password) || g.password_mode==0) /// no password or always ask
  {  
         getstring("Set p3 password for encrypting image\nWarning:keep a record of the password\nor image will be unrecoverable", g.p3password, 1024, 78);
  }

  //// The code will check if password is correct. 
  //// We will encrypt the dimensions and image depth of the image.
  //// Dividing by 27 is important to reduce the chance of
  //// the key leaking.
  //// seed is a global
 
  seed = passwordtoseed(g.p3password, 3);
  salt ^= simplerand() / 27;        //// simplerand changes seed
  seed += salt;                     //// new salted seed
  code ^= simplerand() / 27;        //// simplerand changes seed

  for(i=0; i<1024; i++)
         tempstring[i] ^= simplerand() % 256; //// xor

  //// first 1024 bytes are for salt and image params
  //// second 1024 bytes are reserved for future hashed password
  //// no palette, it is always the same

  fwrite(&salt, 8, 1, fp);              //// 8 byte salt
  fwrite(&code, 8, 1, fp);              //// 8 byte code
  fwrite(tempstring, 1024, 1, fp);      //// image params

  //// write stream cipher of image
  for(j=0;j<z[ci].ysize;j++)
  {  
     for(i=0,i2=0; i<z[ci].xsize; i++,i2+=g.off[bpp])
     {   v = pixelat(z[ci].image[0][j]+i2, bpp);
         v ^= simplerand() / 27; //// xor
         fwrite(&v, sizeof(int), 1, fp);  
     }
  }     
  close_file(fp);  
  return OK;
}


//-------------------------------------------------------------------------//
// readp4file                                                              //
// improved version of p3 encryption                                       //
//-------------------------------------------------------------------------//
int readp4file(char *filename)
{   
   filename = filename;
   return OK;
}

//-------------------------------------------------------------------------//
// writep4file - can only write entire image                               //
// improved version of p3 encryption. Has much higher security and is      //
// designed to make correlation attacks more difficult.                    //
// There is no point in compressing an encrypted file, but the software    //
// doesn't stop you. Please notify me if you find any security weakness.   //
//-------------------------------------------------------------------------//
int writep4file(char *filename)
{
  //// overview---------------------------------------------------------
  //// Use 64 bit random 
  //// Use 16 byte keyword = 128 bits
  //// Encode 4 bytes at a time = 32 bits = 4,294,967,296 possible values
  //// Add counter to avoid coding any repeated pixels
  //// Take random number mod 4,294,967,296
  //// This should give an ambiguity of 2^32 or 4294967296 possibilities 
  ////   for each encoded point and make correlation attack harder
  //// 2^64 =  18,446,744,073,709,551,616
  //// Verified simplerand2 does not repeat for at least 
  //// 2^44            21,434,000,000,000 (1 millionth of max)
  //// 2^44 takes 2 days on Xeon(R) CPU E5-1650 0 @ 3.718 GHz
  //// would take 4383 years to break

  //// description------------------------------------------------------
  //// name        size       written     purpose
  //// -------   -----------   ------  ---------------------
  //// version    8 bytes      yes     not encrypted, format="p4xxxxxx"
  ////                                 where xxxxxx is version no.
  //// s box    256 byte array no      permute image bytes; static
  //// header   1024 bytes     yes     defines x and y size, bpp, 
  ////                                 version, colortype
  //// padding   8 bits        yes     bytes to skip at end
  //// thetime   32 bits       yes     randomize PRNG
  //// nonce     64 bits       yes     prevent correlation attack
  //// salt      32 bits       yes     prevent password collision
  //// llkey     64 bits       no      encrypt bytes, keep hidden, gen. from key
  //// Salt and llkey depend on the password and the image contents.
  //// Nonce depends on time. The idea is that the nonce should never
  //// be repeated, even if the PRNG rolls over.
  //// 
  //// Padding is needed to tell the reader to ignore the extra bytes
  //// at end of file which make file length a multiple of 256.
  //// We add a random number of bytes at the end to conceal it.
  ////
  //// A weakness is due to the one-to-one correspondence between 
  //// plaintext and encoded bytes. An attacker might guess that a 
  //// particular byte corresponds 123 and then use brute force to 
  //// identify the random number at that position. This takes 2^64 / 2^8
  //// = 2^56 guesses, but since there there are 256 possibilities the
  //// search space is still 2^64. On a 3.2 GHz Xeon E5-1650, it took 
  //// me two days just to check 2^44 numbers. Thus, a computer a million 
  //// times faster could find the random number in only two days.
  ////
  //// This is only possible if a one-to-one correspondence exists.
  //// Thus we add an S-box to destroy that correspondence by diffusing 
  //// the information, making the correlation attack impractical.
  //// 
  //// Note that an ordinary bytewise permutation would add no security
  //// because the attacker would still know which byte comes next by looking
  //// at the source code. So the bits must be diffused throughout the
  //// block.
  //// 
  //// Even with the same password and image, the nonce guarantees
  //// that the output is different each time a file is written. 
  //// This is different from p3 format.
  //// 
  //// The password is still the weakest point, so we pad the password 
  //// with 12 bytes of image data. If the user enters 6 char password 
  //// (=36 bits for base 64), the password then becomes 32 + 96 = 
  //// 128 bits. This won't help much if the plaintext is known, but 
  //// in that case there is no need to decode the message at all.
  ////
  //// Sample parameters
  //// 
  //// thetime             2,802,193,898 
  //// nonce   3,913,275,083,763,830,747 
  //// salt   16,447,890,650,910,687,096  
  //// llkey  14,605,540,017,484,196,336  
  //// llseed              1,560,017,778
  ////
  //// We are still limited by the 64 bit integer size. To use bigger ints,
  //// it would be necessary to link to a big number library. That would
  //// mean when (not if) somebody changes the functionality of that
  //// library, it would have to be rewritten. This happened with libjpeg
  //// and libpng already.
  //// -----------------------------------------------------------------

message("Not yet implemented");
return ABORT;

  char tempstring[1024];
  static timeval tt;   
  struct timezone tz;
  int ii, jj=0, ino, bpp;
  uint32_t v=0;    //// must be 32 bit int
  uint32_t thetime;
  FILE *fp=0;
  uint64_t ll=0, llkey, nonce, count, salt=1;
  uint8_t sbox[256], rev_sbox[256];

  ino = ci;
  bpp = z[ino].bpp;

  //// fill unused part of tempstring with junk
  srand((uint32_t)time((time_t *)0));
  memset(tempstring, rand()%0xff, 1024);
  sprintf(tempstring, "%s %s %d %d %d %d %d\n", 
       "p3",
       z[ino].name, 
       z[ino].xsize,
       z[ino].ysize,
       z[ino].bpp,
       z[ino].frames,
       z[ino].colortype);
  tempstring[1024]=0;

  for(ii=0; ii<12; ii++) salt *= max(1, z[ino].image[0][0][ii] & 0xff); 

  if(!strlen(g.p4password) || g.password_mode==0) /// no password or always ask
  {  
         getstring("Set p4 password for encrypting image\nWarning:keep a record of the password\nor image will be unrecoverable", g.p4password, 1024, 78);
  }

  count = 0;
  simplerand2();  
  simplerand2();
  
  gettimeofday(&tt, &tz);
  thetime = (unsigned int)tt.tv_usec;
  thetime ^= simplerand2();         

  //// new llseed based on thetime so sbox is different each time
  llseed = thetime;


printf("sbox\n");
  makesbox(sbox, rev_sbox);

  nonce = simplerand2();
  llkey = (long long unsigned int)passwordtoseed(g.p4password, 3);
  salt ^= simplerand2() % 256;        //// simplerand changes seed
  llseed += salt;                     //// new salted seed from image

  for(ii=0; ii<1024; ii++)  tempstring[ii] ^= simplerand2() % 256; //// xor


//printf("\nthetime  %'u \nnonce  %'llu \nsalt   %'lu  \nllkey  %'llu  \nllseed       %'llu\n", thetime, nonce, salt, llseed, llkey);

// shut up the compiler
filename=filename;count=count;ll=ll;fp=fp;jj=jj;bpp=bpp;v=v;nonce=nonce;
llkey=llkey;

  return OK;
}

//-------------------------------------------------------------------------//
// simplerand2 - random number generator period 2^64                       //
//-------------------------------------------------------------------------//
long long int simplerand2(void)
{

   // arxiv.org/abs/2001.05304
   // xb = (a  xn-1 + c) mod m
   // linear congruential pseudorandom number generator
   // very fast, can do 100 million per second

   static long long int m =  0xffffffffffffffff;  // 2^64 - 1 a power of 2 modulus
                         // 0x10000000000000000 = 2^64
   static long long int a = 1103515245;  // multiplier
   static long long int c = 12345;
   llseed = (a * llseed + c) % m;
   return(llseed);
}

//-------------------------------------------------------------------------//
// makesbox - fill n-sized array with symmetrical lookups                  //
// rijndael s-box                                                          //
//-------------------------------------------------------------------------//
void makesbox(uint8_t *sbox, uint8_t *rev_sbox)
{
     int kk;
     uint8_t p = 1, q = 1;
     
     do {
          p = p ^ (p << 1) ^ (p & 0x80 ? 0x1B : 0);  //// multiply p by 3
          //// divide q by 3 (equals multiplication by 0xf6) 
          q ^= q << 1;
          q ^= q << 2;
          q ^= q << 4;
          q ^= q & 0x80 ? 0x09 : 0;

          //// compute the affine transformation 
          uint8_t xformed = q ^ ROTL8(q, 1) ^ ROTL8(q, 2) ^ ROTL8(q, 3) ^ ROTL8(q, 4);

          sbox[p] = xformed ^ 0x63;
     } while (p != 1);
     sbox[0] = 0x63; //// is a special case since it has no inverse 
     for(kk=0;kk<256;kk++) rev_sbox[sbox[kk]] = kk;
}

