//--------------------------------------------------------------------------//
// xmtnimage69.cc                                                           //
// Manual image registration                                                //
// Latest revision: 04-10-2008                                              //
// See xmtnimage.h for Copyright Notice                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"
#define POINT_MATCH_DELTAS 0   // 0=use raw angles 1=relative angles
                               // 1 works better for large no of points and
                               // is less sensitive to CHECK_POINTS
#define DATASETS 2             // 0=reference 1=unknown 
#define SCALE 32

extern Globals     g;
extern Image      *z;
extern int         ci;
extern int in_warp;
extern int in_warp2d;
int in_manual_registration = 0;
int ino0=0; 
int ino1=0;
int makered=1;
int makegreen=1;
int makeblue=0;
double lthreshold=0, uthreshold=0;
uint lowercolor=0, uppercolor=0;
char warp_filename[1024]="warp.dat";
int reg_warp_gridpoints = 100;

//--------------------------------------------------------------------------//
//  manual_registration                                                     //
//--------------------------------------------------------------------------//
void manual_registration(void)
{
  static Dialog *dialog; 
  static XYData *data;
  static int hit = 0;
  int j,k;
  if(in_manual_registration) return;
  dialog = new Dialog;
  if(dialog==NULL){ message(g.nomemory); return; }
  in_manual_registration = 1;

  if(!hit)     //// Leave data allocated forever for possible reanalysis
  {   data = new XYData[DATASETS];                    //// Don't delete
      for(j=0; j<DATASETS; j++)
      {  
           data[j].title     = new char[FILENAMELENGTH];
           strcpy(data[j].title, "1.data");
           data[j].n         = 0;
           data[j].dims      = 2;
           data[j].ino       = j+1;
           data[j].nmin      = 1;
           data[j].nmax      = RPOINTS;
           data[j].width     = 0;
           data[j].type      = 7;
           data[j].duration  = TEMP;
           data[j].wantpause = 0;
           data[j].wantmessage = 1;
           data[j].win       = 0; // Calculate automatically
      }
      hit = 1;
  } 
  dialog->ptr[4] = (void*)data;

  //// Dialog

  strcpy(dialog->title,"Manual Image Registration");

  dialog->radiono[0]=0;
  for(k=0; k<21; k++) dialog->wcount[k]=2;

  strcpy(dialog->boxes[0],"Ref.image  Unknown image");
  dialog->boxtype[0]=LABEL;

  if(ino0==0) ino0=1;
  if(ino1==0) ino1=2;
  sprintf(dialog->answer[1][0], "Img %d", ino0);
  sprintf(dialog->answer[1][1], "Img %d", ino1);
  strcpy(dialog->boxes[1],"Image no.");
  dialog->boxtype[1]=MULTIPUSHBUTTON; 

  strcpy(dialog->boxes[2],"Select groups of points");
  dialog->boxtype[2]=TOGGLE; 
  dialog->boxset[2]=0;   

  strcpy(dialog->boxes[3],"Cursor movement");
  dialog->boxtype[3]=INTCLICKBOX;
  sprintf(dialog->answer[3][0], "%d", g.warp_cursor);
  dialog->boxmin[3]=1; dialog->boxmax[3]=100;

  strcpy(dialog->boxes[4],"Grid visible");
  dialog->boxtype[4]=TOGGLE;
  dialog->boxset[4] = g.grid_visible;

  strcpy(dialog->boxes[5],"Transparency");
  dialog->boxtype[5]=INTCLICKBOX;
  sprintf(dialog->answer[5][0], "%d", g.warp_transparency);
  dialog->boxmin[5]=0; dialog->boxmax[5]=100;

  strcpy(dialog->boxes[6],"Make ref image green");
  dialog->boxtype[6]=TOGGLE;
  dialog->boxset[6] = makegreen;

  strcpy(dialog->boxes[7],"Make unk image red");
  dialog->boxtype[7]=TOGGLE;
  dialog->boxset[7] = makered; 

  strcpy(dialog->boxes[8],"Capture image");
  dialog->boxtype[8]=PUSHBUTTON;
  dialog->boxset[8] = 0;

  strcpy(dialog->boxes[9],"Read warp vectors");
  dialog->boxtype[9]=PUSHBUTTON;
  dialog->boxset[9] = 0;

  strcpy(dialog->boxes[10],"Save warp vectors");
  dialog->boxtype[10]=PUSHBUTTON;
  dialog->boxset[10] = 0;

  strcpy(dialog->boxes[11],"Apply warp to current image");
  dialog->boxtype[11]=PUSHBUTTON;
  dialog->boxset[11] = 0;

  strcpy(dialog->boxes[12],"Grid size");
  dialog->boxtype[12]=INTCLICKBOX;
  dialog->boxmin[12]=1; dialog->boxmax[12]=200;
  sprintf(dialog->answer[12][0], "%d", reg_warp_gridpoints);

  strcpy(dialog->boxes[13],"Automatic");
  dialog->boxtype[13]=PUSHBUTTON;
  dialog->boxset[13] = 0;

  //// dialog->answer, labels and data are set in registrationcheck()

  dialog->noofradios = 0;
  dialog->noofboxes  = 13;
  dialog->helptopic  = 71;  
  dialog->want_changecicb = 0;
  dialog->f1 = manual_registrationcheck;
  dialog->f2 = null;
  dialog->f3 = null;
  dialog->f4 = manual_registrationfinish;
  dialog->f5 = null;
  dialog->f6 = null;
  dialog->f7 = null;
  dialog->f8 = null;
  dialog->f9 = null;
  dialog->width = 300;
  dialog->height = 0; // calculate automatically
  strcpy(dialog->path,".");
  dialog->transient = 1;
  dialog->wantraise = 0;
  dialog->radiousexy = 0;
  dialog->boxusexy = 0;
  dialog->radioextrawidth = 0;
  dialog->hit = 0;
  dialog->message[0] = 0;      
  dialog->busy = 0;
  dialogbox(dialog);
  return;
}


//--------------------------------------------------------------------------//
//  manual_registrationcheck                                                //
//--------------------------------------------------------------------------//
void manual_registrationcheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  XYData *data = (XYData*)a->ptr[4];
  box=box; boxbutton=boxbutton; radio=radio;
  int dimension=0;
  int k,i,ino,j,count;
  g.getout = 0;
  //// Obtain data
  if(boxbutton >=0) dimension = boxbutton;
  reg_warp_gridpoints = atoi(a->answer[12][0]);
  ino0 = data[0].ino;
  ino1 = data[1].ino;
  if(ino0 >= g.image_count) ino0=ci;
  if(ino1 >= g.image_count) ino1=ci;
  a->boxset[2]=0; // no group pixels
  makegreen = a->boxset[6];
  makered   = a->boxset[7];
  makeblue  = 0; 
  manual_registration_set_ino(a, 0, ino0);
  manual_registration_set_ino(a, 1, ino1);
  switch(box)
  {    case 0: 
               if((z[ino0].colortype != GRAY) && (makegreen || makered || makeblue)) 
                      message("Warning: image is not grayscale\nColors may be wrong");
               switchto(ino1);
               in_warp++;
               warp2d(a, reg_warp_gridpoints, reg_warp_gridpoints, 0, ino0, 
                      makered, makegreen, makeblue);
               if(makegreen | makered | makeblue)
               {   for(k=0; k<g.image_count; k++)   
                   {   z[k].transparency = 0;
                       make_image_gray(k);
                   }
               }
               in_warp--;
               break;
       case 1: a->param[0] = dimension;
               getinteger("Image no.", &data[dimension].ino, 1, g.image_count-1,
               manual_registration_set_ino_ok, null, a, 60);
               break;          
       case 8: // capture image 
               duplicate_img(ci,0,0,g.want_shell,g.window_border);
               message("Image captured");
               change_image_depth(ci,24,1);
               break;
       case 9: // read warp
               readwarp();
               break;
       case 10: // save warp
               savewarp();
               break;
       case 13: // automatically warp
               automatic_registration(ino0, ino1);
               break;
       case 11: // apply warp
               array<int> grid(g.warp_xar+1, g.warp_yar+1, 2);

               //// Must make a copy of image so warping will work
               ino = ci;
               z[ino].transparency = 0;
               duplicate_image(ino, 0, 0, g.want_shell, g.window_border);
               switchto(ci);    // duplicate is new one

               //// Copy warping data into 3d array
               count = 0;
               for(i=0; i<2; i++)
               for(j=0; j<=g.warp_yar; j++)
               for(k=0; k<=g.warp_xar; k++)
               {
                   grid.q[i][j][k] = g.warp_grid[count];
                   count++;
               }

               if(!grid.allocated){ message(g.nomemory); break; }
               if(g.warp_xar==0 || g.warp_yar==0 ||  
                  g.warp_nx==0  || g.warp_ny ==0)
                  { message("Grid not allocated"); break; }
               in_warp++;
               remap_coordinates(ino, 0, ci, grid.q, 0, 0, g.warp_xar, 
                    g.warp_yar, g.warp_nx, g.warp_ny);
               in_warp--;
               repair(ci);
               redraw(ci);
               break;
  }
}


//--------------------------------------------------------------------------//
//  manual_registration_set_ino                                             //
//  set labels in registration and warp dialog box pushbuttons              //
//--------------------------------------------------------------------------//
void manual_registration_set_ino(Dialog *a, int dim, int ino)
{ 
    XYData *data = (XYData*)a->ptr[4];
    sprintf(a->answer[1][dim], "%d", ino);
    set_widget_value(a->boxwidget[1][dim], ino);
    set_widget_label(a->boxwidget[1][dim], a->answer[1][dim]);
    if(dim==0) ino0=ino;
    if(dim==1) ino1=ino;
    data[dim].ino = ino;
}


//--------------------------------------------------------------------------//
//  manual_registration_set_ino_ok                                          //
//--------------------------------------------------------------------------//
void manual_registration_set_ino_ok(clickboxinfo *c)
{
    Dialog *a = (Dialog*) c->client_data;
    int dimension = a->param[0];   
    manual_registration_set_ino(a, dimension, c->answer);
}

//--------------------------------------------------------------------------//
//  manual_registrationfinish                                               //
//--------------------------------------------------------------------------//
void manual_registrationfinish(dialoginfo *a)
{
    a=a;
    if(!in_manual_registration) return;
    g.getout = 1;
    in_manual_registration = 0;
}


//--------------------------------------------------------------------------//
//  highlight_differences                                                   //
//--------------------------------------------------------------------------//
void highlight_differences(int noofargs, char **arg)
{
  noofargs=noofargs; arg=arg;
  static Dialog *dialog; 
  static XYData *data;
  static int hit = 0;
  int j,k;
  dialog = new Dialog;
  if(dialog==NULL){ message(g.nomemory); return; }

  //// Dialog
  if(!hit)     //// Leave data allocated forever for possible reanalysis
  {   data = new XYData[DATASETS];                    //// Don't delete
      for(j=0; j<DATASETS; j++)
      {  
           data[j].title     = new char[FILENAMELENGTH];
           strcpy(data[j].title, "1.data");
           data[j].n         = 0;
           data[j].dims      = 2;
           data[j].ino       = j+1;
           data[j].nmin      = 1;
           data[j].nmax      = RPOINTS;
           data[j].width     = 0;
           data[j].type      = 7;
           data[j].duration  = TEMP;
           data[j].wantpause = 0;
           data[j].wantmessage = 1;
           data[j].win       = 0; // Calculate automatically
      }
      hit = 1;
  } 
  dialog->ptr[4] = (void*)data;

  //// Dialog

  strcpy(dialog->title,"Highlight Image Differences");

  dialog->radiono[0]=0;
  for(k=0; k<21; k++) dialog->wcount[k]=2;

  strcpy(dialog->boxes[0],"Ref.image  Unknown image");
  dialog->boxtype[0]=LABEL;

  if(ino0==0) ino0=1;
  if(ino1==0) ino1=2;
  sprintf(dialog->answer[1][0], "Img %d", ino0);
  sprintf(dialog->answer[1][1], "Img %d", ino1);
  strcpy(dialog->boxes[1],"Image no.");
  dialog->boxtype[1]=MULTIPUSHBUTTON; 

  strcpy(dialog->boxes[2],"Lower ratio threshold");
  dialog->boxtype[2]=STRING; 
  sprintf(dialog->answer[2][0], "%g", lthreshold);

  strcpy(dialog->boxes[3],"Upper ratio threshold");
  dialog->boxtype[3]=STRING; 
  sprintf(dialog->answer[3][0], "%g", uthreshold);

  strcpy(dialog->boxes[4],"Color for decrease");
  dialog->boxtype[4]=STRING; 
  sprintf(dialog->answer[4][0],"%d", lowercolor);

  strcpy(dialog->boxes[5],"Color for increase");
  dialog->boxtype[5]=STRING; 
  sprintf(dialog->answer[5][0],"%d", uppercolor);

  //// dialog->answer, labels and data are set in registrationcheck()

  dialog->noofradios = 0;
  dialog->noofboxes  = 6;
  dialog->helptopic  = 71;  
  dialog->want_changecicb = 0;
  dialog->f1 = highlight_differencescheck;
  dialog->f2 = null;
  dialog->f3 = null;
  dialog->f4 = highlight_differencesfinish;
  dialog->f5 = null;
  dialog->f6 = null;
  dialog->f7 = null;
  dialog->f8 = null;
  dialog->f9 = null;
  dialog->width = 320;
  dialog->height = 0; // calculate automatically
  strcpy(dialog->path,".");
  dialog->transient = 1;
  dialog->wantraise = 0;
  dialog->radiousexy = 0;
  dialog->boxusexy = 0;
  dialog->radioextrawidth = 0;
  dialog->hit = 0;
  dialog->message[0] = 0;      
  dialog->busy = 0;
  dialogbox(dialog);
  return;
}
  

//--------------------------------------------------------------------------//
//  highlight_differencescheck                                              //
//--------------------------------------------------------------------------//
void highlight_differencescheck(dialoginfo *a, int radio, int box, int boxbutton)
{
  XYData *data = (XYData*)a->ptr[4];
  box=box; boxbutton=boxbutton; radio=radio;
  g.getout = 0;

  int dimension=0;
  g.getout = 0;
  //// Obtain data
  if(boxbutton >=0) dimension = boxbutton;
  ino0 = data[0].ino;
  ino1 = data[1].ino;

  lthreshold = atof(a->answer[2][0]);
  uthreshold = atof(a->answer[3][0]);
  lowercolor = atoi(a->answer[4][0]);
  uppercolor = atoi(a->answer[5][0]);
  
  switch(box)
  {    case 0: 
               switchto(ino1);
               in_warp++;
               show_differences(ino0, ino1, lthreshold, uthreshold,
                    lowercolor, uppercolor);
               break;
       case 1: a->param[0] = dimension;
               getinteger("Image no.", &data[dimension].ino, 1, g.image_count-1,
               manual_registration_set_ino_ok, null, a, 60);
               break;          
       case 8: // capture image 
               duplicate_img(ci,0,0,g.want_shell,g.window_border);
               break;
  }
}


//--------------------------------------------------------------------------//
//  highlight_differencesfinish                                             //
//--------------------------------------------------------------------------//
void highlight_differencesfinish(dialoginfo *a)
{
    a=a;
}

//--------------------------------------------------------------------------//
//  show_differences                                                        //
//--------------------------------------------------------------------------//
void show_differences(int ino0, int ino1, double lthreshold, double uthreshold,
                   uint lcolor, uint ucolor)
{
   uchar *address0;
   uchar *address1;
   if((z[ino0].colortype != GRAY) || (z[ino1].colortype != GRAY))
   {   message("Both images must be grayscale");
       return;
   }
   int x,y,x20, x21, v0,v1=0;
   int bpp0 = z[ino0].bpp, bpp1=z[ino1].bpp;
   double ratio=1.0;
   for(y=0; y<z[ino0].ysize; y++)
   {   address0 = z[ino0].image[0][y]; // frame 0
       address1 = z[ino1].image[0][y]; // frame 0
       for(x=0; x<z[ino0].xsize-1; x++)
       {   x20 = x*g.off[bpp0];           
           x21 = x*g.off[bpp1];           
           v0 = pixelat(address0 + x20, bpp0);
           if(between(x, 0, z[ino1].xsize-1) && between(y, 0, z[ino1].ysize-1))
                v1 = pixelat(address1 + x21, 16);
           else v1 = 0;                
           if(v1) ratio=(double)v0/(double)v1; else ratio=1.0;   
           if(ratio < lthreshold) putpixelbytes(address0+x20, lcolor, 1, bpp0, 1);
           else if(ratio > uthreshold) putpixelbytes(address0+x20, ucolor, 1, bpp0, 1);
       }
   }
   rebuild_display(ino0);
   redraw(ino0);
}                   
                   
//--------------------------------------------------------------------------//
//  readwarp                                                                //
//--------------------------------------------------------------------------//
void readwarp(void)
{
   int k,size;
   char temp[256];
   FILE *fp;
   strcpy(warp_filename, getfilename(warp_filename, NULL));
   if(g.escape || strlen(warp_filename)==0){ return; }

   if ((fp=fopen(warp_filename,"r")) == NULL)
   {
        error_message(warp_filename, errno);
   }else
   {   
       fscanf(fp, "%d" ,&g.warp_nx);
       fscanf(fp, "%d" ,&g.warp_ny);
       fscanf(fp, "%d" ,&g.warp_xar);
       fscanf(fp, "%d" ,&g.warp_yar);
       if(!g.warp_grid_allocated)
       {   size = 2 * (g.warp_xar+1) * (g.warp_yar+1);
           g.warp_grid = (int*)malloc(size * sizeof(int));
           g.warp_grid_allocated = 1;
       }
       for(k=0; k<(g.warp_xar+1) * (g.warp_yar+1) * 2; k++)
            fscanf(fp, "%d" ,&g.warp_grid[k]);
       fclose(fp);
       sprintf(temp, "%d x %d warp vectors read",g.warp_nx,g.warp_ny);
       message(temp);
   }
}

//--------------------------------------------------------------------------//
//  savewarp                                                                //
//--------------------------------------------------------------------------//
void savewarp(void)
{
   static PromptStruct ps;
   strcpy(warp_filename, getfilename(warp_filename, NULL));
   if(g.getout || strlen(warp_filename)==0){ g.getout=0; return; }
   ps.filename = warp_filename;
   ps.f1 = savewarp_part2;
   ps.f2 = null;
   ps.ptr = NULL;
   ps.n = (g.warp_xar+1) * (g.warp_yar+1) * 2;
   check_overwrite_file(&ps);
}


 //--------------------------------------------------------------------------//
//  savewarp_part2                                                          //
//--------------------------------------------------------------------------//
void savewarp_part2(PromptStruct *ps)
{
   char temp[FILENAMELENGTH];
   char *filename = ps->filename;
   int k;
   int count = ps->n;
   if(!g.warp_grid_allocated) 
   {    strcpy(temp,"No Data yet\n");
        if(in_warp2d) strcat(temp,"Press Exit to stop warping first,\nthen save");
        message(temp); 
        return;
   }

   FILE *fp;
   if ((fp=fopen(filename,"w")) == NULL)
        error_message(filename, errno);
   else
   {   
       fprintf(fp,"%d\n",g.warp_nx);
       fprintf(fp,"%d\n",g.warp_ny);
       fprintf(fp,"%d\n",g.warp_xar);
       fprintf(fp,"%d\n",g.warp_yar);
       for(k=0; k<count; k++)
            fprintf(fp,"%d\n", g.warp_grid[k]);
        sprintf(temp,"Warp data saved in %s",filename);
        message(temp);
        fclose(fp);
   }
}

//--------------------------------------------------------------------------//
//  automatic_registration - registration of 2 images - not finished.       //
//--------------------------------------------------------------------------//
void automatic_registration(int ino1, int ino2)
{
ino1=ino1;
ino2=ino2;
/*
 printf("a %d %d  %d %d\n",ino1,ino2,g.warp_xar,g.warp_yar);         
    int i,j,k,step;
    double chisq,ochisq,a,b,c,d;
    double f=1;
    in_warp++;
 
    array<int> grid(g.warp_xar, g.warp_yar, 3);
    for(k=0; k<2*g.warp_xar*g.warp_yar; k++)
         grid.data[k] = g.warp_grid[k];
    step =20;
    ochisq = chisquare(ino1, ino2);

    switchto(ino2);
    
while(1)
{ 
    for(j=1; j<g.warp_yar-1; j++)
    for(i=1; i<g.warp_xar-1; i++)
    {
         grid.q[0][j][i] += f*step;
         a = chisquare(ino1, ino2);
         if(a>ochisq) grid.q[0][j][i] -= 2*f*step;


         grid.q[1][j][i] +=  f*step;
         c = chisquare(ino1, ino2);
         if(c>ochisq) grid.q[1][j][i] -=  2*f*step;

        

    }
         chisq = chisquare(ino1, ino2);
         if(chisq<ochisq) f*=0.5; else f*=1.2;
         ochisq = chisq;
 printf("chisq %g %d %d %d %g \n",chisq,grid.q[0][2][2],grid.q[1][2][2],  step,f);         
         in_warp++;
         remap_coordinates(ino1, 0, ino2, grid.q, 0, 0, g.warp_xar-1, 
             g.warp_yar-1, g.warp_nx, g.warp_ny);
         in_warp--;
         repair(ino2);
         redraw(ino2);
         check_event();
         if(g.getout) break;
         if(step==0)break;
 }

 printf("b\n");         
*/
}
